import json
import os
from pathlib import Path
from typing import List, Dict, Union

from ScopeFoundry.generate_loaders_py import generate_loaders_py, get_measurement_name
from .helper_funcs import open_file

IPYNB_DEMO_FNAME = "overview.ipynb"


def new_empty_nb_content(n_cells: int = 2) -> Dict:
    """returns the content of ipynb with n_cell empty cells"""
    return {
        "cells": [
            {
                "cell_type": "code",
                "execution_count": None,
                "metadata": {},
                "outputs": [],
                "source": [],
            }
            for _ in range(n_cells)
        ]
    }


def update_cells(new_sources: List, nb_content: Dict) -> Dict:
    # 1st cell: only update specific lines
    nb_content["cells"][0]["source"] = new_sources[0]
    nb_content["cells"][0]["execution_count"] = None

    # 2nd cell: only update specific lines
    for new_line in new_sources[1]:
        skipp = False
        for old_line in nb_content["cells"][1]["source"]:
            if new_line.strip().lower() in old_line.strip().lower():
                skipp = True
                break
        if not skipp:
            nb_content["cells"][1]["source"].append(new_line)
    nb_content["cells"][1]["execution_count"] = None
    return nb_content


def mk_cell_sources(folder: str) -> List:
    path = Path(folder)

    file_load_lines = []
    fname = "your_file_name.h5"
    for fname in path.rglob("*.h5"):
        mm_name = get_measurement_name(fname)
        file_load_lines.append(f'load_{mm_name}(r"{fname.relative_to(folder)}")')

    src_0 = [
        "# CELL #1: AUTOGENERATED CELL: DO NOT ALTER THIS CELL, PROGRESS MAY BE LOST. use CELL #3 onwards",
        "import numpy as np",
        "from matplotlib import pylab as plt",
        "",
        "",
        f"from h5_data_loaders import load, find_settings",
        "",
        f'data = load(r"{fname}")',
    ]

    src_1 = [
        "# CELL #2 ALTERNATIVE TO CELL #1. Only addition and commenting of lines are persistant",
        "import h5_data_loaders as loaders",
    ] + [f"data = loaders.{a}\n" for a in file_load_lines]

    return [src_0, src_1]


def generate_ipynb(folder: str = ".", ipynb_fname: Union[Path, None] = None) -> Path:
    if ipynb_fname is None:
        ipynb_fname = Path(folder) / IPYNB_DEMO_FNAME

    new_sources = mk_cell_sources(folder)

    nb_content = {}

    if ipynb_fname.exists():
        with open(ipynb_fname, "r") as file:
            nb_content = json.load(file)

    if not "cells" in nb_content or len(nb_content["cells"]) < 2:
        # assuming nb_content is invalid
        nb_content = new_empty_nb_content(2)

    with open(ipynb_fname, "w") as file:
        json.dump(update_cells(new_sources, nb_content), file)

    return ipynb_fname


def analyze_with_ipynb(folder: str = ".") -> None:
    loaders_fname, dset_names = generate_loaders_py(folder)
    ipynb_path = generate_ipynb(folder)

    print("")
    print("generated", loaders_fname, f"with {len(dset_names)} loader(s)")
    print("")
    print("check", ipynb_path)
    print("")

    if ipynb_path.exists():
        open_file(ipynb_path)


if __name__ == "__main__":
    analyze_with_ipynb()



# Form implementation generated from reading ui file '.\raster_scan_base_v2.ui'
#
# Created by: PyQt6 UI code generator 6.9.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets


class Ui_SimpleXYScanWindow(object):
    def setupUi(self, SimpleXYScanWindow):
        SimpleXYScanWindow.setObjectName("SimpleXYScanWindow")
        SimpleXYScanWindow.resize(692, 370)
        SimpleXYScanWindow.setAcceptDrops(False)
        self.centralwidget = QtWidgets.QWidget(parent=SimpleXYScanWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.centralwidget)
        self.gridLayout_2.setVerticalSpacing(3)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.settings_horizontalLayout = QtWidgets.QHBoxLayout()
        self.settings_horizontalLayout.setObjectName("settings_horizontalLayout")
        self.control_groupBox = QtWidgets.QGroupBox(parent=self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.MinimumExpanding,
            QtWidgets.QSizePolicy.Policy.Preferred,
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.control_groupBox.sizePolicy().hasHeightForWidth()
        )
        self.control_groupBox.setSizePolicy(sizePolicy)
        self.control_groupBox.setMaximumSize(QtCore.QSize(16777215, 16777215))
        self.control_groupBox.setObjectName("control_groupBox")
        self.gridLayout = QtWidgets.QGridLayout(self.control_groupBox)
        self.gridLayout.setObjectName("gridLayout")
        self.progressBar = QtWidgets.QProgressBar(parent=self.control_groupBox)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Fixed
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.progressBar.sizePolicy().hasHeightForWidth())
        self.progressBar.setSizePolicy(sizePolicy)
        self.progressBar.setProperty("value", 24)
        self.progressBar.setObjectName("progressBar")
        self.gridLayout.addWidget(self.progressBar, 2, 0, 1, 4)
        self.x_label = QtWidgets.QLabel(parent=self.control_groupBox)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.Expanding,
            QtWidgets.QSizePolicy.Policy.Preferred,
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.x_label.sizePolicy().hasHeightForWidth())
        self.x_label.setSizePolicy(sizePolicy)
        self.x_label.setObjectName("x_label")
        self.gridLayout.addWidget(self.x_label, 1, 4, 1, 2)
        self.dv_label_2 = QtWidgets.QLabel(parent=self.control_groupBox)
        self.dv_label_2.setObjectName("dv_label_2")
        self.gridLayout.addWidget(self.dv_label_2, 6, 4, 1, 1)
        self.h0_doubleSpinBox = QtWidgets.QDoubleSpinBox(parent=self.control_groupBox)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Fixed
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.h0_doubleSpinBox.sizePolicy().hasHeightForWidth()
        )
        self.h0_doubleSpinBox.setSizePolicy(sizePolicy)
        self.h0_doubleSpinBox.setObjectName("h0_doubleSpinBox")
        self.gridLayout.addWidget(self.h0_doubleSpinBox, 4, 2, 1, 1)
        self.Nv_doubleSpinBox = QtWidgets.QDoubleSpinBox(parent=self.control_groupBox)
        self.Nv_doubleSpinBox.setMinimumSize(QtCore.QSize(0, 0))
        self.Nv_doubleSpinBox.setDecimals(0)
        self.Nv_doubleSpinBox.setObjectName("Nv_doubleSpinBox")
        self.gridLayout.addWidget(self.Nv_doubleSpinBox, 6, 7, 1, 1)
        self.dv_doubleSpinBox = QtWidgets.QDoubleSpinBox(parent=self.control_groupBox)
        self.dv_doubleSpinBox.setObjectName("dv_doubleSpinBox")
        self.gridLayout.addWidget(self.dv_doubleSpinBox, 6, 5, 1, 1)
        self.line = QtWidgets.QFrame(parent=self.control_groupBox)
        self.line.setFrameShape(QtWidgets.QFrame.Shape.HLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Shadow.Sunken)
        self.line.setObjectName("line")
        self.gridLayout.addWidget(self.line, 3, 0, 1, 8)
        self.y_doubleSpinBox = QtWidgets.QDoubleSpinBox(parent=self.control_groupBox)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Fixed
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.y_doubleSpinBox.sizePolicy().hasHeightForWidth()
        )
        self.y_doubleSpinBox.setSizePolicy(sizePolicy)
        self.y_doubleSpinBox.setObjectName("y_doubleSpinBox")
        self.gridLayout.addWidget(self.y_doubleSpinBox, 2, 6, 1, 2)
        self.h_label = QtWidgets.QLabel(parent=self.control_groupBox)
        self.h_label.setObjectName("h_label")
        self.gridLayout.addWidget(self.h_label, 4, 0, 1, 1)
        self.Nh_doubleSpinBox = QtWidgets.QDoubleSpinBox(parent=self.control_groupBox)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Fixed
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.Nh_doubleSpinBox.sizePolicy().hasHeightForWidth()
        )
        self.Nh_doubleSpinBox.setSizePolicy(sizePolicy)
        self.Nh_doubleSpinBox.setMinimumSize(QtCore.QSize(50, 0))
        self.Nh_doubleSpinBox.setFrame(True)
        self.Nh_doubleSpinBox.setDecimals(0)
        self.Nh_doubleSpinBox.setMaximum(1000.0)
        self.Nh_doubleSpinBox.setProperty("value", 1000.0)
        self.Nh_doubleSpinBox.setObjectName("Nh_doubleSpinBox")
        self.gridLayout.addWidget(self.Nh_doubleSpinBox, 4, 7, 1, 1)
        self.v_label_2 = QtWidgets.QLabel(parent=self.control_groupBox)
        self.v_label_2.setObjectName("v_label_2")
        self.gridLayout.addWidget(self.v_label_2, 6, 0, 1, 1)
        self.v_axis_comboBox = QtWidgets.QComboBox(parent=self.control_groupBox)
        self.v_axis_comboBox.setObjectName("v_axis_comboBox")
        self.gridLayout.addWidget(self.v_axis_comboBox, 6, 1, 1, 1)
        self.start_pushButton = QtWidgets.QPushButton(parent=self.control_groupBox)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Fixed
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.start_pushButton.sizePolicy().hasHeightForWidth()
        )
        self.start_pushButton.setSizePolicy(sizePolicy)
        self.start_pushButton.setObjectName("start_pushButton")
        self.gridLayout.addWidget(self.start_pushButton, 0, 4, 1, 4)
        self.v1_doubleSpinBox = QtWidgets.QDoubleSpinBox(parent=self.control_groupBox)
        self.v1_doubleSpinBox.setObjectName("v1_doubleSpinBox")
        self.gridLayout.addWidget(self.v1_doubleSpinBox, 6, 3, 1, 1)
        self.h1_doubleSpinBox = QtWidgets.QDoubleSpinBox(parent=self.control_groupBox)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Fixed
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.h1_doubleSpinBox.sizePolicy().hasHeightForWidth()
        )
        self.h1_doubleSpinBox.setSizePolicy(sizePolicy)
        self.h1_doubleSpinBox.setObjectName("h1_doubleSpinBox")
        self.gridLayout.addWidget(self.h1_doubleSpinBox, 4, 3, 1, 1)
        self.dv_label = QtWidgets.QLabel(parent=self.control_groupBox)
        self.dv_label.setObjectName("dv_label")
        self.gridLayout.addWidget(self.dv_label, 4, 4, 1, 1)
        self.y_label = QtWidgets.QLabel(parent=self.control_groupBox)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.Expanding,
            QtWidgets.QSizePolicy.Policy.Preferred,
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.y_label.sizePolicy().hasHeightForWidth())
        self.y_label.setSizePolicy(sizePolicy)
        self.y_label.setObjectName("y_label")
        self.gridLayout.addWidget(self.y_label, 1, 6, 1, 2)
        self.dh_doubleSpinBox = QtWidgets.QDoubleSpinBox(parent=self.control_groupBox)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Fixed
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.dh_doubleSpinBox.sizePolicy().hasHeightForWidth()
        )
        self.dh_doubleSpinBox.setSizePolicy(sizePolicy)
        self.dh_doubleSpinBox.setObjectName("dh_doubleSpinBox")
        self.gridLayout.addWidget(self.dh_doubleSpinBox, 4, 5, 1, 1)
        self.label_2 = QtWidgets.QLabel(parent=self.control_groupBox)
        self.label_2.setObjectName("label_2")
        self.gridLayout.addWidget(self.label_2, 4, 6, 1, 1)
        self.x_doubleSpinBox = QtWidgets.QDoubleSpinBox(parent=self.control_groupBox)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Fixed
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.x_doubleSpinBox.sizePolicy().hasHeightForWidth()
        )
        self.x_doubleSpinBox.setSizePolicy(sizePolicy)
        self.x_doubleSpinBox.setObjectName("x_doubleSpinBox")
        self.gridLayout.addWidget(self.x_doubleSpinBox, 2, 4, 1, 2)
        self.v0_doubleSpinBox = QtWidgets.QDoubleSpinBox(parent=self.control_groupBox)
        self.v0_doubleSpinBox.setObjectName("v0_doubleSpinBox")
        self.gridLayout.addWidget(self.v0_doubleSpinBox, 6, 2, 1, 1)
        self.h_axis_comboBox = QtWidgets.QComboBox(parent=self.control_groupBox)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Fixed
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.h_axis_comboBox.sizePolicy().hasHeightForWidth()
        )
        self.h_axis_comboBox.setSizePolicy(sizePolicy)
        self.h_axis_comboBox.setObjectName("h_axis_comboBox")
        self.gridLayout.addWidget(self.h_axis_comboBox, 4, 1, 1, 1)
        self.label_3 = QtWidgets.QLabel(parent=self.control_groupBox)
        self.label_3.setObjectName("label_3")
        self.gridLayout.addWidget(self.label_3, 6, 6, 1, 1)
        self.scan_type_comboBox = QtWidgets.QComboBox(parent=self.control_groupBox)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Fixed
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.scan_type_comboBox.sizePolicy().hasHeightForWidth()
        )
        self.scan_type_comboBox.setSizePolicy(sizePolicy)
        self.scan_type_comboBox.setObjectName("scan_type_comboBox")
        self.gridLayout.addWidget(self.scan_type_comboBox, 0, 0, 1, 4)
        self.label = QtWidgets.QLabel(parent=self.control_groupBox)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.Expanding,
            QtWidgets.QSizePolicy.Policy.Preferred,
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label.sizePolicy().hasHeightForWidth())
        self.label.setSizePolicy(sizePolicy)
        self.label.setObjectName("label")
        self.gridLayout.addWidget(self.label, 1, 0, 1, 3)
        self.settings_horizontalLayout.addWidget(self.control_groupBox)
        self.verticalLayout_3 = QtWidgets.QVBoxLayout()
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.groupBox = QtWidgets.QGroupBox(parent=self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.Maximum, QtWidgets.QSizePolicy.Policy.Minimum
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.groupBox.sizePolicy().hasHeightForWidth())
        self.groupBox.setSizePolicy(sizePolicy)
        self.groupBox.setMinimumSize(QtCore.QSize(100, 0))
        self.groupBox.setObjectName("groupBox")
        self.gridLayout_3 = QtWidgets.QGridLayout(self.groupBox)
        self.gridLayout_3.setObjectName("gridLayout_3")
        self.continuous_scan_checkBox = QtWidgets.QCheckBox(parent=self.groupBox)
        self.continuous_scan_checkBox.setObjectName("continuous_scan_checkBox")
        self.gridLayout_3.addWidget(self.continuous_scan_checkBox, 0, 0, 1, 1)
        self.save_h5_checkBox = QtWidgets.QCheckBox(parent=self.groupBox)
        self.save_h5_checkBox.setObjectName("save_h5_checkBox")
        self.gridLayout_3.addWidget(self.save_h5_checkBox, 1, 0, 1, 1)
        self.verticalLayout_3.addWidget(self.groupBox)
        self.groupBox_2 = QtWidgets.QGroupBox(parent=self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Preferred
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.groupBox_2.sizePolicy().hasHeightForWidth())
        self.groupBox_2.setSizePolicy(sizePolicy)
        self.groupBox_2.setMinimumSize(QtCore.QSize(100, 0))
        self.groupBox_2.setObjectName("groupBox_2")
        self.gridLayout_4 = QtWidgets.QGridLayout(self.groupBox_2)
        self.gridLayout_4.setObjectName("gridLayout_4")
        self.show_previous_scans_checkBox = QtWidgets.QCheckBox(parent=self.groupBox_2)
        self.show_previous_scans_checkBox.setObjectName("show_previous_scans_checkBox")
        self.gridLayout_4.addWidget(self.show_previous_scans_checkBox, 0, 0, 1, 1)
        self.clear_previous_scans_pushButton = QtWidgets.QPushButton(
            parent=self.groupBox_2
        )
        self.clear_previous_scans_pushButton.setObjectName(
            "clear_previous_scans_pushButton"
        )
        self.gridLayout_4.addWidget(self.clear_previous_scans_pushButton, 1, 0, 1, 1)
        self.verticalLayout_3.addWidget(self.groupBox_2)
        self.settings_horizontalLayout.addLayout(self.verticalLayout_3)
        self.details_groupBox = QtWidgets.QGroupBox(parent=self.centralwidget)
        self.details_groupBox.setObjectName("details_groupBox")
        self.details_layout = QtWidgets.QVBoxLayout(self.details_groupBox)
        self.details_layout.setObjectName("details_layout")
        self.settings_horizontalLayout.addWidget(self.details_groupBox)
        self.device_details_GroupBox = QtWidgets.QGroupBox(parent=self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.Preferred,
            QtWidgets.QSizePolicy.Policy.Preferred,
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.device_details_GroupBox.sizePolicy().hasHeightForWidth()
        )
        self.device_details_GroupBox.setSizePolicy(sizePolicy)
        self.device_details_GroupBox.setObjectName("device_details_GroupBox")
        self.device_details_layout = QtWidgets.QGridLayout(self.device_details_GroupBox)
        self.device_details_layout.setObjectName("device_details_layout")
        self.settings_horizontalLayout.addWidget(self.device_details_GroupBox)
        self.gridLayout_2.addLayout(self.settings_horizontalLayout, 0, 0, 1, 1)
        self.plot_groupBox = QtWidgets.QGroupBox(parent=self.centralwidget)
        self.plot_groupBox.setTitle("")
        self.plot_groupBox.setFlat(True)
        self.plot_groupBox.setObjectName("plot_groupBox")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.plot_groupBox)
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_2.setSpacing(0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.gridLayout_2.addWidget(self.plot_groupBox, 1, 0, 1, 1)
        SimpleXYScanWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(parent=SimpleXYScanWindow)
        self.menubar.setEnabled(True)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 692, 17))
        self.menubar.setObjectName("menubar")
        SimpleXYScanWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(parent=SimpleXYScanWindow)
        self.statusbar.setEnabled(True)
        self.statusbar.setObjectName("statusbar")
        SimpleXYScanWindow.setStatusBar(self.statusbar)

        self.retranslateUi(SimpleXYScanWindow)
        QtCore.QMetaObject.connectSlotsByName(SimpleXYScanWindow)
        SimpleXYScanWindow.setTabOrder(self.start_pushButton, self.h0_doubleSpinBox)
        SimpleXYScanWindow.setTabOrder(self.h0_doubleSpinBox, self.h1_doubleSpinBox)
        SimpleXYScanWindow.setTabOrder(self.h1_doubleSpinBox, self.dh_doubleSpinBox)
        SimpleXYScanWindow.setTabOrder(self.dh_doubleSpinBox, self.Nh_doubleSpinBox)
        SimpleXYScanWindow.setTabOrder(self.Nh_doubleSpinBox, self.v0_doubleSpinBox)
        SimpleXYScanWindow.setTabOrder(self.v0_doubleSpinBox, self.v1_doubleSpinBox)
        SimpleXYScanWindow.setTabOrder(self.v1_doubleSpinBox, self.dv_doubleSpinBox)
        SimpleXYScanWindow.setTabOrder(self.dv_doubleSpinBox, self.Nv_doubleSpinBox)

    def retranslateUi(self, SimpleXYScanWindow):
        _translate = QtCore.QCoreApplication.translate
        SimpleXYScanWindow.setWindowTitle(
            _translate("SimpleXYScanWindow", "Simple XY Scan")
        )
        self.control_groupBox.setTitle(_translate("SimpleXYScanWindow", "Scan Control"))
        self.x_label.setText(_translate("SimpleXYScanWindow", "current H"))
        self.dv_label_2.setText(_translate("SimpleXYScanWindow", "dv"))
        self.h_label.setText(_translate("SimpleXYScanWindow", "H"))
        self.v_label_2.setText(_translate("SimpleXYScanWindow", "V"))
        self.start_pushButton.setText(_translate("SimpleXYScanWindow", "Start"))
        self.dv_label.setText(_translate("SimpleXYScanWindow", "dh"))
        self.y_label.setText(_translate("SimpleXYScanWindow", "current V"))
        self.label_2.setText(_translate("SimpleXYScanWindow", "Nh"))
        self.label_3.setText(_translate("SimpleXYScanWindow", "Nv"))
        self.label.setText(_translate("SimpleXYScanWindow", "Progress"))
        self.groupBox.setTitle(_translate("SimpleXYScanWindow", "Data"))
        self.continuous_scan_checkBox.setText(
            _translate("SimpleXYScanWindow", "Continuous")
        )
        self.save_h5_checkBox.setText(_translate("SimpleXYScanWindow", "Save H5"))
        self.groupBox_2.setTitle(_translate("SimpleXYScanWindow", "View"))
        self.show_previous_scans_checkBox.setText(
            _translate("SimpleXYScanWindow", "Show \n" "Prev Scans")
        )
        self.clear_previous_scans_pushButton.setText(
            _translate("SimpleXYScanWindow", "Clear Prev")
        )
        self.details_groupBox.setTitle(_translate("SimpleXYScanWindow", "Details"))
        self.device_details_GroupBox.setTitle(
            _translate("SimpleXYScanWindow", "Device Details")
        )



from dataclasses import dataclass
from ScopeFoundry.cb32_uuid import cb32_uuid


import time
from datetime import datetime
from pathlib import Path


@dataclass
class DatasetMetadata:
    unique_id: str  # persistent identifier for dataset
    u: str
    t0: int
    h5_file_path: Path
    app_name: str = None

    def get_file_path(self, suffix: str = ".h5") -> Path:
        return self.h5_file_path.with_suffix(suffix)

    def savefig_kwargs(self, suffix: str = ".png") -> dict:
        return {
            "filename_or_obj": self.get_file_path(suffix),
            "metadata": {
                "Title": self.h5_file_path.stem,
                # "Author": self.unique_id,
                "Description": self.unique_id,
                "Created": datetime.fromtimestamp(self.t0).strftime(
                    "%a, %d %b %Y %H:%M:%S GMT"
                ),
                "Software": "ScopeFoundry",
                # "Disclaimers": "This dataset is for internal use only.",
                # "Warning": "This dataset is not for public distribution.",
                "Source": self.app_name,    
                # "Comment": "This dataset is for internal use only.",
            },
        }


def new_dataset_metadata(measurement=None, fname: str = None) -> DatasetMetadata:
    unique_id, u = cb32_uuid()  # persistent identifier for dataset
    t0 = time.time()

    if measurement is not None:
        app = measurement.app
        if fname is None:
            fname = app.settings["data_fname_format"].format(
                app=app,
                measurement=measurement,
                timestamp=datetime.fromtimestamp(t0),
                unique_id=unique_id,
                unique_id_short=unique_id[0:13],
                ext="h5",
            )
        h5_file_path = Path(app.settings["save_dir"]) / fname
        return DatasetMetadata(unique_id, u, t0, h5_file_path, app_name=app.name)

    if fname is None:
        h5_file_path = Path.cwd() / f"{datetime.fromtimestamp(t0):%y%m%d_%H%M%S}.h5"
    else:
        h5_file_path = Path(fname).with_suffix(".h5")

    return DatasetMetadata(unique_id, u, t0, h5_file_path, app_name=None)



from typing import Any, Callable, Dict, Iterable, List, Tuple, Union

from ScopeFoundry.base_app import BaseMicroscopeApp

# Each actuator can be defined with a tuple of settings paths. The following formats are supported:
# 1. (name, position_path | position_read_func, target_position_path | target_position_write_func)
# 2. (name, target_position_path | target_position_write_func) -> target_position_path=position_path
# 3. (target_position_path) -> name=target_position_path=position_path

WriteFunc = Callable[[Any], None]
ReadFunc = Callable[[], Any]
WriteInfo = Union[str, WriteFunc]
ReadInfo = Union[str, ReadFunc, None]
ActuatorFuncs = Tuple[ReadFunc, WriteFunc]
ActuatorInfos = Tuple[ReadInfo, WriteInfo]


ActuatorDefinitions = Union[
    str,
    Tuple[str],
    Tuple[str, WriteInfo],
    Tuple[str, WriteInfo, ReadInfo],
]


def add_all_possible_actuators_and_parse_definitions(
    actuator_definitions: Iterable[ActuatorDefinitions],
    app: BaseMicroscopeApp,
) -> Dict[str, ActuatorInfos]:
    ds = list(actuator_definitions)
    ds.extend(app.get_setting_paths(filter_has_hardware_write=True))
    return parse_definitions(ds)


def parse_definitions(
    actuator_definitions: Iterable[ActuatorDefinitions],
) -> Dict[str, ActuatorInfos]:
    """returns a list of tuples with the actuator name, read path and write path"""
    d = {}
    for defs in actuator_definitions:
        label = defs[0]
        if isinstance(defs, str):
            d[label] = (None, defs)
        elif len(defs) == 1:
            d[label] = (None, defs[0])
        if len(defs) == 2:
            d[label] = (None, defs[1])
        elif len(defs) == 3:
            d[label] = (defs[1], defs[2])
    return d


def get_actuator_funcs(
    app: BaseMicroscopeApp,
    actuator_definitions: Dict[str, ActuatorInfos],
) -> Dict[str, ActuatorFuncs]:
    return {
        k: mk_actuator_func(app, read_info=v[0], write_info=v[1])
        for k, v in actuator_definitions.items()
    }


def mk_actuator_func(
    app: BaseMicroscopeApp,
    read_info: ReadInfo,
    write_info: WriteInfo,
) -> ActuatorFuncs:

    if callable(write_info):
        write_func = write_info
    else:
        write_func = app.get_lq(write_info).update_value

    if read_info is None:
        read_func = lambda: 0
    elif callable(read_info):
        read_func = read_info
    else:
        read_func = app.get_lq(read_info).read_from_hardware

    return read_func, write_func



from qtpy import QtCore, QtGui, QtWidgets


class LoggingWidget(QtWidgets.QWidget):

    def __init__(
        self,
        parent: QtWidgets.QWidget = None,
        buffer_len=500,
    ) -> None:
        super().__init__(parent)
        self.buffer_len = buffer_len
        self.messages = []

        self.setWindowTitle("Log")
        self.setLayout(QtWidgets.QVBoxLayout())
        self.search_lineEdit = QtWidgets.QLineEdit()
        self.text_edit = QtWidgets.QTextEdit("")

        self.layout().addWidget(self.search_lineEdit)
        self.layout().addWidget(self.text_edit)

        self.text_edit.document().setDefaultStyleSheet("body{font-family: Courier;}")
        self.search_lineEdit.editingFinished.connect(self.on_search)

    @QtCore.Slot(str)
    def on_new_log(self, log_entry: str):
        self.messages.append(log_entry)
        if len(self.messages) > self.buffer_len:
            self.messages = ["...<br>"] + self.messages[-self.buffer_len :]
        self.text_edit.setHtml("\n".join(self.messages))
        self.text_edit.moveCursor(QtGui.QTextCursor.End)

    @QtCore.Slot()
    def on_search(self):
        search_text = self.search_lineEdit.text().lower()
        messages = [message for message in self.messages if search_text in message.lower()]
        self.text_edit.setHtml("\n".join(messages))
        self.text_edit.moveCursor(QtGui.QTextCursor.End)



import os
from pathlib import Path
import subprocess


def publish_hw(
    github_username,
    subfolder,
    subfolder_path,
    message="Initial commit",
    private_or_public="--public",
):
    print(f"Processing folder: {subfolder}")

    # Initialize a Git repository if not already initialized

    if not os.path.exists(os.path.join(subfolder_path, ".git")):
        subprocess.run(["git", "init"], cwd=subfolder_path)

    subprocess.run(["git", "branch", "-m", "main"], cwd=subfolder_path)
    # Add all files and commit
    subprocess.run(["git", "add", "."], cwd=subfolder_path)
    subprocess.run(["git", "commit", "-m", message], cwd=subfolder_path)

    # Create a GitHub repository using GitHub CLI
    repo_name = subfolder
    subprocess.run(
        [
            "gh",
            "repo",
            "create",
            f"{github_username}/HW_{repo_name}",
            private_or_public,
            "--source",
            subfolder_path,
            "--push",
        ],
        cwd=subfolder_path,
    )
    # Push the local repository to GitHub

    subprocess.run(
        [
            "git",
            "remote",
            "add",
            "origin",
            f"https://github.com/{github_username}/HW_{repo_name}.git",
        ],
        cwd=subfolder_path,
    )

    subprocess.run(["git", "push", "-u", "origin", "main"], cwd=subfolder_path)

    print(f"Published {subfolder} as a GitHub repository.")


def publish_subfolders_as_repos(parent_dir, github_username):
    """
    Publishes each subfolder in the given directory as a separate GitHub repository.

    Parameters:
    - parent_dir: Path to the parent directory containing subfolders.
    - github_username: Your GitHub username.
    """
    for subfolder in os.listdir(parent_dir):
        subfolder_path = os.path.join(parent_dir, subfolder)

        # Skip if not a directory
        if not os.path.isdir(subfolder_path):
            continue

        publish_hw(github_username, subfolder, subfolder_path)


if __name__ == "__main__":
    parent_directory = Path(__file__).parent  # Adjust this to your parent directory
    gh_username = "ubene"  # Replace with your GitHub username
    publish_subfolders_as_repos(parent_directory, gh_username)



# intended to make a new setup by running:
# python -m ScopeFoundry.tools.new_setup

import re
import shutil
from pathlib import Path


def convert_import_statement(import_line: str) -> str:
    pattern = r"from ScopeFoundry\.examples\.((?:\w+\.)*\w+) imp(\w+)"
    replacement = r"from \1 imp\2"
    new_import = re.sub(pattern, replacement, import_line)
    return new_import


def convert_import_statements(input_file_path: str) -> None:
    with open(input_file_path, "r") as infile:
        lines = infile.readlines()

    with open(input_file_path, "w") as outfile:
        for line in lines:
            outfile.write(convert_import_statement(line))


def copy_scopefoundry_examples() -> None:

    source_dir = Path(__file__).resolve().parent.parent.parent / "examples"
    dest_dir = Path.cwd()

    if not source_dir.exists():
        print(f"Source directory {source_dir} does not exist.")
        return

    for file_path in source_dir.rglob("*"):
        if file_path.is_file():
            if file_path.suffix not in (".py", ".ini", ".json"):
                continue
            relative_path = file_path.relative_to(source_dir)
            dest_file_path = dest_dir / relative_path

            if dest_file_path.exists():
                print(f"Skipped {dest_file_path}, already exists.")
                continue

            dest_file_path.parent.mkdir(parents=True, exist_ok=True)

            shutil.copy2(file_path, dest_file_path)
            convert_import_statements(dest_file_path)
            print(f"Copied {file_path} to {dest_file_path} and adjusted imports")


def new_app() -> str:
    copy_scopefoundry_examples()
    print(
        """
test installation by running:
    
    python fancy_app.py

"""
    )
    return "python fancy_app.py"


if __name__ == "__main__":
    new_app()



import sys
from pathlib import Path
from typing import Dict

from qtpy import QtGui, QtWidgets

from ScopeFoundry import ini_io
from ScopeFoundry.tools.page import Page
from ScopeFoundry.tools.pages.new_app import NewApp
from ScopeFoundry.tools.pages.new_hardware import NewHardware
from ScopeFoundry.tools.pages.new_measurement import NewMeasurement
from ScopeFoundry.tools.pages.publish_hw import PublishHW
from ScopeFoundry.tools.pages.welcome import WelcomePage

ROOT = Path(__file__).parent.parent
ICONS_PATH = ROOT / "base_app/icons"
DEFAULT_INI_FILE_CONTENT = {
    "new hardware/authors": "Benedikt Ursprung",
    "publish HW on GitHub/gh_username": "UBene",
    "publish HW on GitHub/ScopeFoundryHW_dir": "/Users/benediktursprung/Library/CloudStorage/OneDrive-Personal/scope_foundries/scopefoundry_trovatello_lab/ScopeFoundry/tools",
    "publish HW on GitHub/private_or_public": "--public",
}


class ToolsApp(QtWidgets.QMainWindow):

    def __init__(self) -> None:
        super().__init__()

        self.tab_widget = QtWidgets.QTabWidget()
        self.setCentralWidget(self.tab_widget)
        self.pages: Dict[str, Page] = {}
        self.pages_index = []

    def add_page(self, page: Page) -> None:
        page.setup()
        page.setup_figure()
        self.tab_widget.addTab(page.ui, page.name)
        self.pages[page.name] = page
        self.pages_index.append(page.name)

    def show_page(self, name: str) -> None:
        self.tab_widget.setCurrentIndex(self.pages_index.index(name))

    def read_and_set_settings(self) -> None:
        ins = self.read_settings()
        print(ins)
        for path, value in ins.items():
            d, name = path.split("/")
            if name == "ScopeFoundryHW_dir":
                if not Path(value).is_dir():
                    continue
            self.pages[d].settings[name] = value

    def read_settings(self):
        if not Path("tools_defaults.ini").is_file():
            ini_io.save_settings("tools_defaults.ini", DEFAULT_INI_FILE_CONTENT)
            return DEFAULT_INI_FILE_CONTENT
        DEFAULT_INI_FILE_CONTENT.update(ini_io.load_settings("tools_defaults.ini"))
        return DEFAULT_INI_FILE_CONTENT

    def save_settings(self):
        ins = self.read_settings()
        for path, value in ins.items():
            d, name = path.split("/")

            ins[path] = self.pages[d].settings[name]

        ini_io.save_settings("tools_defaults.ini", ins)


def main() -> None:
    if len(sys.argv) > 1:
        start_app(" ".join(sys.argv[1:]))
    else:
        start_app()


def start_app(page="Welcome") -> None:
    app = QtWidgets.QApplication([])
    window = ToolsApp()
    window.setWindowTitle("ScopeFoundry tools")
    logo_icon = QtGui.QIcon(str(ICONS_PATH / "scopefoundry_logo2C_1024.png"))
    window.setWindowIcon(logo_icon)
    window.add_page(WelcomePage(window))
    window.add_page(NewApp(window))
    window.add_page(NewHardware(window))
    window.add_page(NewMeasurement(window))
    window.add_page(PublishHW(window))
    window.show()
    window.show_page(page)
    window.read_settings()
    window.read_and_set_settings()
    app.exec_()
    window.save_settings()
    sys.exit()


if __name__ == "__main__":
    main()



from ScopeFoundry import HardwareComponent
from ScopeFoundry.scanning import BaseRaster3DSlowScanV2


class Example3DSlowScanMeasure(BaseRaster3DSlowScanV2):

    name = "example_3d_scan"

    def scan_specific_setup(self):
        self.detector: HardwareComponent = self.app.hardware["noiser_200"]

    def pre_scan_setup(self):
        if self.settings["save_h5"]:
            self.signal_map = self.h5_meas_group.create_dataset(
                name="signal_map", shape=self.scan_shape, dtype=float
            )

    def collect_pixel(self, pixel_num, k, j, i):
        signal = self.detector.settings.get_lq("signal").read_from_hardware()
        self.display_image_map[k, j, i] = signal
        if self.settings["save_h5"]:
            self.signal_map[k, j, i] = signal



import numpy as np


class LQ3Vector:

    def __init__(self, x_lq, y_lq, z_lq):
        self.x_lq = x_lq
        self.y_lq = y_lq
        self.z_lq = z_lq

    @property
    def values(self):
        return np.array([self.x_lq.val, self.y_lq.val, self.z_lq.val])

    @property
    def length(self):
        vec = self.values
        return np.sqrt(np.dot(vec, vec))

    @property
    def normed_values(self):
        return self.values / self.length

    def dot(self, _lq_vector):
        """
        returns the scalar product with _lq_vector
        """
        return np.dot(_lq_vector.values, self.values)

    def project_on(self, _lq_vector):
        return np.dot(_lq_vector.normed_values, self.values)

    def angle_to(self, _lq_vector):
        return np.arccos(np.dot(_lq_vector.normed_values, self.normed_values))

    def add_listener(self, func, argtype=(), **kwargs):
        self.x_lq.add_listener(func, argtype, **kwargs)
        self.y_lq.add_listener(func, argtype, **kwargs)
        self.z_lq.add_listener(func, argtype, **kwargs)



from qtpy import QtWidgets
import h5py

from ScopeFoundry.data_browser import DataBrowserView


class H5TreeView(DataBrowserView):

    name = "h5_tree"

    def is_file_supported(self, fname):
        return ".h5" in fname

    def setup(self):
        self.ui = QtWidgets.QTextEdit("file_info")

    def on_change_data_filename(self, fname=None):
        self.ui.setText(f"loading {fname}")
        try:
            self.tree_str = f"{fname}\n{'=' * len(fname)}\n"
            with h5py.File(fname, "r") as file:
                file.visititems(self._visitfunc)
            self.ui.setText(self.tree_str)

        except Exception as err:
            self.databrowser.ui.statusbar.showMessage(
                "failed to load %s:\n%s" % (fname, err)
            )
            raise (err)

    def _visitfunc(self, name, node):

        level = len(name.split("/"))
        indent = "    " * level
        localname = name.split("/")[-1]

        if isinstance(node, h5py.Group):
            self.tree_str += f"{indent}|> {localname}\n"
        elif isinstance(node, h5py.Dataset):
            self.tree_str += f"{indent}|D {localname}: {node.shape} {node.dtype}\n"
        for key, val in node.attrs.items():
            self.tree_str += f"{indent}    |- {key} = {val}\n"


class H5TreeSearchView(DataBrowserView):

    name = "h5_tree_search"

    def is_file_supported(self, fname):
        return ".h5" in fname

    def setup(self):

        # self.settings.New('search_text', dtype=str, initial="")

        self.ui = QtWidgets.QWidget()
        self.ui.setLayout(QtWidgets.QVBoxLayout())
        self.search_lineEdit = QtWidgets.QLineEdit()
        self.tree_textEdit = QtWidgets.QTextEdit("")

        self.ui.layout().addWidget(self.search_lineEdit)
        self.ui.layout().addWidget(self.tree_textEdit)

        # self.settings.search_text.connect_to_widget(self.search_lineEdit)
        # self.settings.search_text.add_listener(self.on_new_search_text)
        self.search_text = ""
        self.fname = None

        self.search_lineEdit.textChanged.connect(self.on_new_search_text)

    def on_change_data_filename(self, fname=None):
        self.tree_textEdit.setText(f"loading {fname}")
        try:
            self.fname = fname
            self.on_new_search_text()
            self.databrowser.ui.statusbar.showMessage("")

        except Exception as err:
            msg = "Failed to load %s:\n%s" % (fname, err)
            self.databrowser.ui.statusbar.showMessage(msg)
            self.tree_textEdit.setText(msg)
            raise (err)

    def on_new_search_text(self, x=None):
        if self.fname is None:
            return
        if x is not None:
            self.search_text = x.lower()
        old_scroll_pos = self.tree_textEdit.verticalScrollBar().value()
        self.tree_str = ""

        with h5py.File(self.fname, "r") as file:
            file.visititems(self._visitfunc)

        self.tree_text_html = f"""<html><b>{self.fname}</b><hr/>
        <div style="font-family: Courier;">
         {self.tree_str} 
         </div>
         </html>"""

        self.tree_textEdit.setText(self.tree_text_html)
        self.tree_textEdit.verticalScrollBar().setValue(old_scroll_pos)

    def _visitfunc(self, name, node):

        level = len(name.split("/"))
        indent = "&nbsp;" * 4 * (level - 1)

        # indent = '<span style="color:blue;">'.format(level*4)
        localname = name.split("/")[-1]

        # search_text = self.settings['search_text'].lower()
        search_text = self.search_text
        if search_text and (search_text in localname.lower()):
            localname = f"""<span style="color: red;">{localname}</span>"""

        if isinstance(node, h5py.Group):
            self.tree_str += f"{indent}|> <b>{localname}/</b><br/>"
        elif isinstance(node, h5py.Dataset):
            self.tree_str += (
                f"{indent}|D <b>{localname}</b>: {node.shape} {node.dtype}<br/>"
            )
        for key, val in node.attrs.items():
            if search_text:
                if search_text in str(key).lower():
                    key = f"""<span style="color: red;">{key}</span>"""
                if search_text in str(val).lower():
                    val = f"""<span style="color: red;">{key}</span>"""
            self.tree_str += (
                f"{indent}&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;|- <i>{key}</i> = {val}<br/>"
            )



import logging
import sys
import time

import numpy as np

from ScopeFoundry import BaseMicroscopeApp
from ScopeFoundry.examples.ScopeFoundryHW.bsinc_noiser200 import Noiser200HW
from ScopeFoundry.examples.ScopeFoundryHW.simulon_xyz_stage import SimulonXYZStageHW
from ScopeFoundry.scanning import (
    BaseRaster2DFrameSlowScan,
    BaseRaster2DSlowScan,
    BaseRaster2DSlowScanV2,
    BaseRaster3DSlowScan,
    BaseRaster3DSlowScanV2,
)

logging.basicConfig(level=logging.INFO)
logging.getLogger("ScopeFoundry").setLevel(logging.DEBUG)


class TestRaster2DSlowScan(BaseRaster2DSlowScan):
    name = "test_cart_2d_slow_scan"

    def scan_specific_setup(self):
        self.stage = self.app.hardware["simulon_xyz_stage"]
        self.detector = self.app.hardware["noiser_200"]

    def move_position_start(self, h, v):
        self.stage.settings["x_target_position"] = h
        self.stage.settings["y_target_position"] = v

    def move_position_fast(self, h, v, dh, dv):
        return self.move_position_start(h, v)

    def move_position_slow(self, h, v, dh, dv):
        return self.move_position_start(h, v)

    def pre_scan_setup(self):
        self.display_update_period = 0.050  # seconds

        if self.settings["save_h5"]:
            self.test_data = self.h5_meas_group.create_dataset(
                "test_data", self.scan_shape, dtype=float
            )

        self.prev_px = time.time()

    def post_scan_cleanup(self):
        print("post_scan_cleanup")

    def collect_pixel(self, pixel_i, k, j, i):
        # print pixel_i, k,j,i
        t0 = time.time()
        # px_data = np.random.rand()
        # px_data = t0 - self.prev_px
        x0, y0 = self.pos
        x_set = self.stage.settings["x_position"]
        y_set = self.stage.settings["y_position"]
        x_hw = self.stage.settings.get_lq("x_position").read_from_hardware()
        y_hw = self.stage.settings.get_lq("y_position").read_from_hardware()
        if np.abs(x_hw - x0) > 1:
            self.log.debug("=" * 60)
            self.log.debug("pos      {} {}".format(x0, y0))
            self.log.debug("settings {} {}".format(x_set, y_set))
            self.log.debug("hw       {} {}".format(x_hw, y_hw))
            self.log.debug("settings value delta {} {}".format(x_set - x0, y_set - y0))
            self.log.debug("read_hw  value delta {} {}".format(x_hw - x0, y_hw - y0))
            self.log.debug("=" * 60)

        x = x_hw
        y = y_hw

        px_data = (
            np.sinc((x - 50) * 0.05) ** 2 * np.sinc(0.05 * (y - 50)) ** 2
        )  # + 0.05*np.random.random()
        # px_data = (x-xhw)**2 + ( y-yhw)**2
        # if px_data > 1:
        #    print('hw', x, xhw, y, yhw)
        self.display_image_map[k, j, i] = px_data
        if self.settings["save_h5"]:
            self.test_data[k, j, i] = px_data
        time.sleep(self.settings["pixel_time"])
        # self.prev_px = t0


class TestRaster2DFrameSlowScan(BaseRaster2DFrameSlowScan):
    name = "test_cart_2d_frame_slow_scan"

    def scan_specific_setup(self):
        self.stage = self.app.hardware["simulon_xyz_stage"]
        self.detector = self.app.hardware["noiser_200"]

    def pre_scan_setup(self):  # called at the begining of run
        self.display_update_period = 0.050  # seconds

        if self.settings["save_h5"]:
            self.test_data = self.create_h5_framed_dataset(
                "test_data", self.display_image_map
            )

        self.prev_px = time.time()

    def post_scan_cleanup(self):  # called at the end of run
        print("post_scan_cleanup")

    def collect_pixel(self, pixel_i, frame_i, k, j, i):
        t0 = time.time()
        x_hw = self.stage.settings.get_lq("x_position").read_from_hardware()
        y_hw = self.stage.settings.get_lq("y_position").read_from_hardware()

        theta = np.pi / 10.0 * frame_i
        x = x_hw * np.cos(theta) - y_hw * np.sin(theta)
        y = x_hw * np.sin(theta) + y_hw * np.cos(theta)

        px_data = (
            np.sinc((x - 50) * 0.05) ** 2 * np.sinc(0.05 * (y - 50)) ** 2
        )  # + 0.05*np.random.random()

        self.display_image_map[k, j, i] = px_data
        if self.settings["save_h5"]:
            self.test_data[frame_i, k, j, i] = px_data
        time.sleep(self.settings["pixel_time"])

    def on_new_frame(self, frame_i):
        print("on_new_frame")
        if self.settings["save_h5"]:
            self.extend_h5_framed_dataset(self.test_data, frame_i)

    def on_end_frame(self, frame_i):
        pass


class TestRaster3DSlowScan(BaseRaster3DSlowScan):
    name = "test_cart_3d_slow_scan"

    def scan_specific_setup(self):
        self.stage = self.app.hardware["simulon_xyz_stage"]
        self.detector = self.app.hardware["noiser_200"]

    def move_position_start(self, h, v, z):
        self.stage.settings["x_target_position"] = h
        self.stage.settings["y_target_position"] = v
        self.stage.settings["z_target_position"] = z

    def move_position_slow(self, h, v, dh, dv):
        self.stage.settings["x_target_position"] = h
        self.stage.settings["y_target_position"] = v

    def pre_scan_setup(self):
        self.display_update_period = 0.050  # seconds

        if self.settings["save_h5"]:
            self.test_data = self.h5_meas_group.create_dataset(
                "test_data", self.scan_shape, dtype=float
            )

        self.prev_px = time.time()

    def post_scan_cleanup(self):
        print("post_scan_cleanup")

    def collect_pixel(self, pixel_i, k, j, i):

        t0 = time.time()

        x0, y0 = self.pos
        x_set = self.stage.settings["x_position"]
        y_set = self.stage.settings["y_position"]

        x_hw = self.stage.settings.get_lq("x_position").read_from_hardware()
        y_hw = self.stage.settings.get_lq("y_position").read_from_hardware()
        z_hw = self.stage.settings.get_lq("z_position").read_from_hardware()

        if np.abs(x_hw - x0) > 1:
            self.log.debug("=" * 60)
            self.log.debug("pos      {} {}".format(x0, y0))
            self.log.debug("settings {} {}".format(x_set, y_set))
            self.log.debug("hw       {} {}".format(x_hw, y_hw))
            self.log.debug("settings value delta {} {}".format(x_set - x0, y_set - y0))
            self.log.debug("read_hw  value delta {} {}".format(x_hw - x0, y_hw - y0))
            self.log.debug("=" * 60)

        x = x_hw
        y = y_hw

        px_data = np.sinc((x - 50) * 0.05) ** 2 * np.sinc(0.05 * (y - 50)) ** 2

        # NOTE: In 3D index k represents z position / while in 2D k represents 'frame'
        self.display_image_map[k, j, i] = px_data
        if self.settings["save_h5"]:
            self.test_data[k, j, i] = px_data
        time.sleep(self.settings["pixel_time"])


class Example2DSlowScanMeasure(BaseRaster2DSlowScanV2):

    name = "example_2d_scan"

    def scan_specific_setup(self):
        self.detector = self.app.hardware["noiser_200"]

    def pre_scan_setup(self):
        if self.settings["save_h5"]:
            self.signal_map = self.h5_meas_group.create_dataset(
                name="signal_map", shape=self.scan_shape, dtype=float
            )

    def collect_pixel(self, pixel_num, k, j, i):
        signal = self.detector.settings.get_lq("signal").read_from_hardware()
        self.display_image_map[k, j, i] = signal
        if self.settings["save_h5"]:
            self.signal_map[k, j, i] = signal


class Example3DSlowScanMeasure(BaseRaster3DSlowScanV2):

    name = "example_3d_scan"

    def scan_specific_setup(self):
        self.detector = self.app.hardware["noiser_200"]

    def pre_scan_setup(self):
        if self.settings["save_h5"]:
            self.signal_map = self.h5_meas_group.create_dataset(
                name="signal_map", shape=self.scan_shape, dtype=float
            )

    def collect_pixel(self, pixel_num, k, j, i):
        signal = self.detector.settings.get_lq("signal").read_from_hardware()
        self.display_image_map[k, j, i] = signal
        if self.settings["save_h5"]:
            self.signal_map[k, j, i] = signal


class TestRaster2DSlowScanApp(BaseMicroscopeApp):

    name = "app"

    def setup(self):

        stage = self.add_hardware(SimulonXYZStageHW(self))
        stage.settings["connected"] = True

        detector = self.add_hardware(Noiser200HW(self))
        detector.settings["connected"] = True

        self.add_measurement(
            TestRaster2DSlowScan(self, h_limits=(0, 100), v_limits=(0, 100))
        )
        self.add_measurement(
            TestRaster2DFrameSlowScan(self, h_limits=(0, 100), v_limits=(0, 100))
        )

        self.add_measurement(
            TestRaster3DSlowScan(
                self, h_limits=(0, 100), v_limits=(0, 100), z_limits=(-1, 1)
            )
        )

        for m in self.measurements.values():
            S = m.settings
            # S = self.test_cart_2d_slow_scan.settings
            S["h1"] = 100
            S["v1"] = 100
            S["dh"] = S["dv"] = 1.0

        actuators = (
            (
                "x_position",
                "hw/simulon_xyz_stage/x_position",
                "hw/simulon_xyz_stage/x_target_position",
            ),
            (
                "y_position",
                "hw/simulon_xyz_stage/y_position",
                "hw/simulon_xyz_stage/y_target_position",
            ),
            (
                "z_position",
                "hw/simulon_xyz_stage/z_position",
                "hw/simulon_xyz_stage/z_target_position",
            ),
        )

        self.add_measurement(
            Example2DSlowScanMeasure(
                self, actuators=actuators, h_limits=(0, 100), v_limits=(0, 100)
            )
        )
        self.add_measurement(
            Example3DSlowScanMeasure(
                self,
                actuators=actuators,
                h_limits=(0, 100),
                v_limits=(0, 100),
                z_limits=(-1, 1),
            )
        )


if __name__ == "__main__":
    app = TestRaster2DSlowScanApp([])

    # from ScopeFoundry.flask_web_view.flask_web_view import MicroscopeFlaskWebThread
    # app.flask_thread = MicroscopeFlaskWebThread(app)
    # app.flask_thread.start()

    sys.exit(app.exec_())



from dataclasses import dataclass, field
from typing import Callable, Dict, List, Tuple

from qtpy import QtCore, QtGui, QtWidgets

from ScopeFoundry.dynamical_widgets.tools import Tools
from ScopeFoundry.helper_funcs import filter_with_patterns


class OperationsQObject(QtCore.QObject):

    added = QtCore.Signal(str)
    removed = QtCore.Signal(str)


@dataclass
class Operation:
    name: str
    func: Callable
    description: str = field(default="")
    icon_path: str = field(default="")
    path: str = field(default="")

    def new_button(self) -> QtWidgets.QPushButton:
        op_button = QtWidgets.QPushButton(self.name)
        op_button.setObjectName(self.name)
        op_button.clicked.connect(lambda checked, f=self.func: f())
        op_button.setToolTip(f"<b>{self.path}</b><p><p>{self.description}")
        if self.icon_path:
            op_button.setIcon(QtGui.QIcon(str(self.icon_path)))
        return op_button


class Operations:

    def __init__(self, path="") -> None:
        self.q_object = OperationsQObject()
        self._widgets_managers_: List[OperationWidgetsManager] = []
        self._operations: Dict[str, Operation] = {}
        self.path = path

    def new(
        self, name: str, func: Callable, description: str = "", icon_path: str = ""
    ) -> Operation:
        return self.add(
            Operation(name, func, description, icon_path, f"{self.path}/{name}")
        )

    def add(self, operation: Operation) -> Operation:
        assert operation.name not in self._operations
        self._operations[operation.name] = operation
        self.q_object.added.emit(operation.name)
        return operation

    def remove(self, name: str) -> None:
        if name not in self._operations:
            return
        del self._operations[name]
        self.q_object.removed.emit(name)

    def get(self, name: str) -> Operation:
        return self._operations[name]

    def new_button(self, name: str) -> QtWidgets.QPushButton:
        return self._operations[name].new_button()

    def keys(self) -> List[str]:
        return self._operations.keys()

    def __contains__(self, k: str) -> bool:
        return self._operations.__contains__(k)

    def __iter__(self) -> iter:
        return self._operations.__iter__()

    # For backwards compatbility, returning op_func rather the Operation objects.
    def items(self) -> List[Tuple[str, Callable]]:
        return ((k, v.func) for k, v in self._operations.items())

    def __getitem__(self, key: str) -> Callable:
        return self._operations[key].func


class OperationWidgetsManager:

    def __init__(self, operations: Operations, tools: Tools) -> None:
        self.operations = operations
        self.tools = tools

        operations.q_object.added.connect(self.add)
        operations.q_object.removed.connect(self.remove)

        self.widgets = {}
        self.update()

    def add(self, name: str) -> None:
        self.update()

    def remove(self, name: str) -> None:
        widget = self.widgets.pop(name, None)
        if widget is None:
            return
        self.tools.remove_from_layout(name, widget)

    def update(self) -> None:
        for op_name in filter_with_patterns(
            self.operations.keys(), self.tools.include, self.tools.exclude
        ):
            if op_name in self.widgets.keys():
                continue
            op_button = self.operations.new_button(op_name)
            self.widgets[op_name] = op_button
            self.tools.add_to_layout(None, op_button)



from ScopeFoundry import HardwareComponent
from ScopeFoundry.scanning import BaseRaster2DSlowScanV2


class Example2DSlowScanMeasure(BaseRaster2DSlowScanV2):

    name = "example_2d_scan"

    def scan_specific_setup(self):
        self.detector: HardwareComponent = self.app.hardware["noiser_200"]

    def pre_scan_setup(self):
        if self.settings["save_h5"]:
            self.signal_map = self.h5_meas_group.create_dataset(
                name="signal_map", shape=self.scan_shape, dtype=float
            )

    # def setup_figure(self):
    #     super().setup_figure()
    #     add_to_layout(
    #         self.detector,
    #         self.ui.device_details_layout,
    #         include=("connected", "signal", "voltage"),
    #     )

    def collect_pixel(self, pixel_num, k, j, i):
        signal = self.detector.settings.get_lq("signal").read_from_hardware()
        self.display_image_map[k, j, i] = signal
        if self.settings["save_h5"]:
            self.signal_map[k, j, i] = signal



"""
Created on Tue Apr  1 09:25:48 2014
@author: esbarnard
"""

import sys
import threading
import time
import traceback
from pathlib import Path
from typing import Any, Callable, Dict, Union
from warnings import warn

from qtpy import QtCore, QtGui, QtWidgets


from .base_app import BaseMicroscopeApp
from .dynamical_widgets.generic_widget import add_to_layout, new_widget
from .dynamical_widgets.tree_widget import SubtreeManager
from .helper_funcs import (
    get_child_path,
    get_logger_from_class,
    init_docs_path,
    load_qt_ui_file,
    itemize_launchers,
)
from .logged_quantity import LQCollection
from .operations import Operations
from .dataset_metadata import new_dataset_metadata


class MeasurementQThread(QtCore.QThread):
    def __init__(self, measurement, parent=None):
        super(MeasurementQThread, self).__init__(parent)
        self.measurement = measurement

    def run(self):
        self.measurement._thread_run()


class Measurement:
    """
    Base class for ScopeFoundry Measurement objects

    to subclass, implement :meth:`setup`, :meth:`run`

    for measurements with graphical interfaces,
    subclass and additionally implement :meth:`setup_figure`, :meth:`update_display`


    Run States:

    stop_first -> run_starting -> run_pre_run --> run_thread_starting --> run_thread_run -->

    run_thread_end --> run_post_run --> stop_success | stop_interrupted | stop_failure

    """

    def __init__(self, app: BaseMicroscopeApp, name: Union[str, None] = None):

        self.q_object = MeasurementQObject(self)
        self.measurement_sucessfully_completed = (
            self.q_object.measurement_sucessfully_completed
        )
        self.measurement_interrupted = self.q_object.measurement_interrupted

        self.log = get_logger_from_class(self)

        if not hasattr(self, "name"):
            self.name = self.__class__.__name__

        if name is not None:
            self.name = name

        self.app = app

        self.display_update_period = 0.1  # seconds

        self.acq_thread = None

        self.interrupt_measurement_called = False

        self.settings = LQCollection(path=f"mm/{self.name}")
        self.operations = Operations(path=f"mm/{self.name}")
        self._subtree_managers_ = []
        self._widgets_managers_ = []

        self.activation = self.settings.New(
            name="activation",
            dtype=bool,
            ro=False,
            description=f"<i>{self.name}</i>",
            protected=True,
        )  # does the user want to the thread to be running

        self.run_state = self.settings.New(
            "run_state", dtype=str, initial="stop_first", ro=True, protected=True
        )
        self.progress = self.settings.New(
            "progress", dtype=float, unit="%", si=False, ro=True, protected=True
        )
        self.settings.New(
            "profile",
            dtype=bool,
            initial=False,
            description="Run a profile on the run to find performance problems",
        )

        self.activation.updated_value[bool].connect(self.start_stop)

        self.add_operation("start", self.start)
        self.add_operation("interrupt", self.interrupt)
        # self.add_operation('terminate', self.terminate)
        # self.add_operation("setup", self.setup)
        # self.add_operation("setup_figure", self.setup_figure)
        self.add_operation("update_display", self.update_display)
        self.add_operation(
            name="show_ui",
            op_func=self.show_ui,
            icon_path=self.app.icons_path / "show_ui_logo.png",
        )
        self.add_operation("Reload_Code", self.reload_code)

        if hasattr(self, "ui_filename"):
            self.load_ui()
        self.setup()
        self.subwin = None  # will be set when the measurement is added to the app

    def setup(self):
        """Override this to set up logged quantities and gui connections
        Runs during __init__, before the hardware connection is established
        Should generate desired LoggedQuantities"""
        pass
        # raise NotImplementedError()

    def setup_figure(self):
        """
        Override setup_figure to build graphical interfaces.
        This function is run on ScopeFoundry startup.
        """
        self.log.info("Empty setup_figure called")
        pass

    def start(self):
        """
        Starts the measurement

        calls *pre_run*
        creates acquisition thread
        runs thread
        starts display timer which calls update_display periodically
        calls post run when thread is finished
        """
        # self.start_stop(True)
        self.activation.update_value(True)

    def _start(self):
        """
        INTERNAL DO NOT CALL DIRECTLY

        Starts the measurement

        calls *pre_run*
        creates acquisition thread
        runs thread
        starts display timer which calls update_display periodically
        connects a signal/slot that calls post run when thread is finished
        """
        self.interrupt_measurement_called = False
        self.run_state.update_value("run_starting")

        msg = f"measurement {self.name} start called from thread: {repr(threading.get_ident())}"
        self.log.info(msg)

        if self.is_thread_alive():
            msg = f"Cannot start a new measurement while still measuring {self.acq_thread} {self.is_measuring()}"
            raise RuntimeError(msg)

        # remove previous qthread with delete later
        # if self.acq_thread is not None:
        #    self.acq_thread.deleteLater()
        self.acq_thread = MeasurementQThread(self)
        self.acq_thread.finished.connect(self._call_post_run)
        # self.measurement_state_changed.emit(True)
        # self.running.update_value(True)
        self.run_state.update_value("run_prerun")
        try:
            self.pre_run()
        except Exception as err:
            # print("err", err)
            self.run_state.update_value("stop_failure")
            self.activation.update_value(False)
            raise

        self.run_state.update_value("run_thread_starting")
        self.acq_thread.start()
        self.run_state.update_value("run_thread_run")
        self.t_start = time.time()
        self.q_object.display_update_timer.start(int(self.display_update_period * 1000))

    def pre_run(self):
        """Override this method to enable main-thread initialization prior to measurement thread start"""
        pass

    def run(self):
        """
        *run* method runs in an separate thread and is used for data acquisition

        No GUI updates should occur within the *run* function, any Qt related GUI work
        should occur in :meth:`update_display`

        Don't call this directly!
        """

        if hasattr(self, "_run"):
            self.log.warning("warning _run is deprecated, use run")
            self._run()
        else:
            raise NotImplementedError(f"Measurement {self.name}.run() not defined")

    def _call_post_run(self):
        """
        Don't call this directly!
        """
        self.run_state.update_value("run_post_run")
        try:
            self.post_run()
        except Exception as err:
            raise
        finally:
            self.activation.update_value(False)
            self.run_state.update_value(self.end_state)

    def post_run(self):
        """Override this method to enable main-thread finalization after to measurement thread completes"""
        pass

    def _thread_run(self):
        """
        This function governs the behavior of the measurement thread.
        """
        print(self.name, "_thread_run thread_id:", threading.get_ident())
        self.progress.update_value(50.0)  # default 50% w/o time remaining estimation
        self._t0 = time.time()
        try:
            if self.settings["profile"]:
                import cProfile

                profile = cProfile.Profile()
                profile.enable()
            self.run()
            success = True
        except Exception as err:
            success = False
            raise
        finally:
            self.run_state.update_value("run_thread_end")

            # self.running.update_value(False)
            self.set_progress(0.0)  # set progress bars back to zero
            # self.measurement_state_changed.emit(False)
            if self.interrupt_measurement_called:
                self.measurement_interrupted.emit()
                self.interrupt_measurement_called = False
                end_state = "stop_interrupted"
            elif not success:
                end_state = "stop_failure"
            else:
                self.measurement_sucessfully_completed.emit()
                end_state = "stop_success"
            if self.settings["profile"]:
                profile.disable()
                profile.print_stats(sort="time")

            self.end_state = end_state

    @property
    def gui(self):
        self.log.warning(
            "Measurement.gui is deprecated, use Measurement.app "
            + repr(DeprecationWarning)
        )
        return self.app

    def set_progress(self, pct):
        """
        This function updates the logged quantity progress which is used for the display of progress bars in the UI.

        ==============  ==============================================================================================
        **Arguments:**
        pct             The percentage of progress given by a measurement module
        ==============  ==============================================================================================
        """
        self.progress.update_value(pct)

        if pct:
            text = f"{self.name} (in {to_etr_str((100 - pct)/pct * (time.time() - self._t0))})"
        else:
            text = self.name

        if hasattr(self, "subwin"):
            self.subwin.setWindowTitle(text)

        for manager in self._subtree_managers_:
            manager.set_header_text(0, text, None)

    def _interrupt(self):
        """
        Kindly ask the measurement to stop.

        This raises the :attr:`interrupt_measurement_called` flag
        To actually stop, the threaded :meth:`run` method must check
        for this flag and exit
        """
        self.log.info(f"measurement {self.name} stopping {self.settings['run_state']}")
        # print("{self.name} interrupt(): run_state={ self.settings['run_state']}")
        if self.settings["run_state"].startswith("run"):
            self.log.info(f"measurement {self.name} interrupt called")
            self.interrupt_measurement_called = True
        # self.activation.update_value(False)
        # Make sure display is up to date
        # self.q_object._on_display_update_timer()

    def interrupt(self):
        self.activation.update_value(False)

    def terminate(self):
        """
        Terminate MeasurementQThread. Usually a bad idea:
        This will not clean up the thread correctly and usually
        requires a reboot of the App
        """
        self.acq_thread.terminate()

    def start_stop(self, start):
        """
        Use boolean *start* to either start (True) or
        interrupt (False) measurement. Test.
        """
        self.log.info(f"{self.name} start_stop {start}")
        if start:
            self._start()
        else:
            self._interrupt()

    def is_measuring(self):
        """
        Returns whether the acquisition thread is running
        """
        # print(self.name, "is_measuring run_state", self.settings['run_state'])
        return self.settings["run_state"].startswith("run")

        """
        if self.acq_thread is None:
            self.running.update_value(False)
            self.activation.update_value(False)
            self.settings['progress'] = 0.0
            return False
        else:
            #resp =  self.acq_thread.is_alive()
            resp = self.acq_thread.isRunning()
            self.running.update_value(resp)            
            return resp
        """

    def is_thread_alive(self):
        if self.acq_thread is None:
            return False
        else:
            # resp =  self.acq_thread.is_alive()
            resp = self.acq_thread.isRunning()
            return resp

    def update_display(self):
        "Override this function to provide figure updates when the display timer runs"
        pass

    def add_logged_quantity(self, name, **kwargs):
        """
        Create a new :class:`LoggedQuantity` and adds it to the measurement's
        :attr:`settings` (:class:`LQCollection`)
        """
        lq = self.settings.New(name=name, **kwargs)
        return lq

    def start_nested_measure_and_wait(
        self,
        measure,
        nested_interrupt: bool = True,
        polling_func: Callable = None,
        polling_time: float = 0.1,
    ):
        """
        Start another nested measurement *measure* and wait until completion.
        Should be called with run function.


        Optionally it can call a polling function *polling_func* with no arguments
        at an interval *polling_time* in seconds.

        if *nested_interrupt* is True then interrupting the nested *measure* will
        also interrupt the outer measurement. *nested_interrupt* defaults to True

        returns True if successful run, otherwise returns false for a run failure or interrupted measurement
        """

        self.log.info(
            f"Starting nested measurement {measure.name} from {self.name} on thread id {threading.get_ident()}"
        )

        measure.interrupt_measurement_called = False
        measure.start()

        # Wait until measurement has started, timeout of 1 second
        t0 = time.time()
        while not measure.is_measuring():
            time.sleep(0.010)
            if time.time() - t0 > 1.0:
                print(
                    f"{self.name}: nested measurement {measure.name} has not started before timeout"
                )
                return measure.settings["run_state"] == "stop_success"

        last_polling = time.time()

        # Now that it is running, wait until done
        while measure.is_measuring():
            if self.interrupt_measurement_called:
                # print('nest outer interrupted', self.interrupt_measurement_called)
                measure.interrupt()

            if measure.interrupt_measurement_called and nested_interrupt:
                # THIS IS MAYBE UNSAFE???: measure.interrupt_measurement_called might be also TRUE if measure finished successfully?
                # IDEA to TEST: also check the measure.settings['run_state'].startswidth('stop')
                print(
                    "nested interrupt bubbling up",
                    measure.interrupt_measurement_called,
                    self.interrupt_measurement_called,
                )
                self.interrupt()

            time.sleep(0.010)

            # polling
            t = time.time()
            if (
                measure.settings["run_state"] == "run_thread_run"
                and polling_func
                and t - last_polling > polling_time
            ):
                try:
                    polling_func()
                except Exception as err:
                    self.log.error(
                        f"start_nested_measure_and_wait polling failed {err}"
                    )
                last_polling = t

        # returns True if successful run, otherwise,
        # returns false for a run failure or interrupted measurement
        return measure.settings["run_state"] == "stop_success"

    def load_ui(self, ui_fname=None):
        """
        Loads and shows user interface.

        ==============  ===============================================================
        **Arguments:**
        ui_fname        filename of user interface file (usually made with Qt Designer)
        ==============  ===============================================================
        """

        # TODO destroy and rebuild UI if it already exists
        if ui_fname is not None:
            self.ui_filename = ui_fname
        # Load Qt UI from .ui file
        self.ui = load_qt_ui_file(self.ui_filename)
        # self.show_ui()

    def show_ui(self):
        """
        Shows the graphical user interface of this measurement. :attr:`ui`
        """
        self.app.bring_measure_ui_to_front(self)

    def new_control_widgets(self, title: str = None, include=None, exclude=None):
        """creates scroll area group box that updates on dynamical add/remove of settings/operations"""
        if title is None:
            title = self.name
        return new_widget(self, title, include, exclude)

    def add_to_layout(self, layout, include=None, exclude=None):
        add_to_layout(self, layout, include, exclude)

    def new_start_stop_button(self, texts=("▶  START", "🛑  STOP")):
        btn = QtWidgets.QPushButton()
        btn.setObjectName("start_stop_button")
        self.activation.connect_to_pushButton(btn, (), texts, "")
        if len(texts[0]) <= 1 and len(texts[1]) <= 1:
            btn.setFixedWidth(22)
        return btn

    def add_operation(
        self, name: str, op_func: Callable[[], None], description="", icon_path=""
    ):
        """
        Create an operation for the Measurement.

        *op_func* is a function that will be called upon operation activation

        operations are typically exposed in the default ScopeFoundry gui via a pushButton

        :type name: str
        :type op_func: QtCore.Slot or Callable without Argument
        :type description: str
        """
        self.operations.new(name, op_func, description, icon_path)

    def remove_operation(self, name: str):
        self.operations.remove(name)

    def web_ui(self):
        return f"Hardware {self.name}"

    def reload_code(self):
        import inspect
        import sys

        if sys.version_info[1] <= 11:
            import xreload
        else:
            from ScopeFoundry import xreload

        mod = inspect.getmodule(self)

        x = xreload.xreload(mod)
        print("Reloaded code", x)

    def on_new_subtree(self, subtree: SubtreeManager):
        progress_bar = QtWidgets.QProgressBar()
        progress_bar.setObjectName("measurement_progress")
        self.progress.connect_to_widget(progress_bar)

        btn = self.new_start_stop_button(texts=("▶", "🛑"))
        btn.clicked.connect(self.show_ui)

        widget = QtWidgets.QWidget()
        layout = QtWidgets.QHBoxLayout(widget)
        layout.addWidget(progress_bar)
        layout.addWidget(btn)
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)
        subtree.tree_widget.setItemWidget(subtree.header_item, 1, widget)

        subtree.progress_bar = progress_bar

    @property
    def tree_progressBar(self):
        # for backward compatibility
        warn(
            "Measurement.tree_progressBar deprecated, loop through Measurement._subtree_managers_[i].progress_bar instead.",
            DeprecationWarning,
        )
        return self._subtree_managers_[0].progress_bar

    def on_right_click(self):
        cmenu = QtWidgets.QMenu()
        cmenu.addAction(self.name).setEnabled(False)
        cmenu.addSeparator()

        cmenu.addAction("Start", self.start)
        cmenu.addAction("Interrupt", self.interrupt)
        ui = self.app.load_measure_ui(self)
        if ui is not None:
            cmenu.addAction("Show", self.show_ui)

        init_docs_path(self.docs_path, self.settings)
        pairs = itemize_launchers(self.docs_path, self.app.launch_browser, False)
        if pairs:
            cmenu.addSeparator()
            for name, func in pairs:
                cmenu.addAction(name, func)

        cmenu.exec_(QtGui.QCursor.pos())

    @property
    def docs_path(self):
        # need to get derived class directory and name
        module = sys.modules[self.__class__.__module__]
        if self.__class__.__module__ == "__main__":
            self.log.warning(
                f"Measurement.docs_path called from __main__, place {self.__class__} in a separate file"
            )
            return get_child_path(self) / "docs"
        return Path(module.__file__).parent / f"{module.__file__.strip('.py')}_docs"

    def new_dataset_metadata(self, fname=None):
        """Creates and returns Measurement.dataset_metadata

        Use Measurement.dataset_metadata for consistent (h5, csv, png with persistent indentifyer)
        """
        self.dataset_metadata = new_dataset_metadata(measurement=self, fname=fname)
        return self.dataset_metadata

    def save_h5(self, data: Dict[str, Any], dataset_metadata=None):
        """
        opens new a .h5 file, stores *data*, and closes the file.

        If *dataset_metadata* is None, Measurement.new_dataset_metadata() is called which updates
        Measurement.dataset_metadata
        """
        self.open_new_h5_file(data=data, dataset_metadata=dataset_metadata)
        self.close_h5_file()

    def open_new_h5_file(self, data: Dict[str, Any] = None, dataset_metadata=None):
        """
        sets attributes:
          - Measurement.h5_file
          - Measurement.h5_meas_group

        If *dataset_metadata* is None, Measurement.new_dataset_metadata() is called which updates
        Measurement.dataset_metadata

        returns Measurement.h5_meas_group
        """

        # there might be a file open already
        self.close_h5_file()

        if dataset_metadata is None:
            dataset_metadata = self.new_dataset_metadata()

        from ScopeFoundry import h5_io

        self.h5_file = h5_io.h5_base_file(self.app, dataset_metadata=dataset_metadata)
        self.h5_meas_group = h5_io.h5_create_measurement_group(self, self.h5_file)

        if data:
            for name, value in data.items():
                try:
                    _ = iter(value)
                except TypeError:
                    # not iterable
                    self.h5_meas_group.attrs[name] = value
                else:
                    # iterable
                    self.h5_meas_group.create_dataset(name=name, data=value)

        return self.h5_meas_group

    def close_h5_file(self):
        if hasattr(self, "h5_file") and self.h5_file.id is not None:
            self.h5_file.close()


class MeasurementQObject(QtCore.QObject):

    measurement_sucessfully_completed = QtCore.Signal(())
    """signal sent when full measurement is complete"""
    measurement_interrupted = QtCore.Signal(())
    """signal sent when  measurement is complete due to an interruption"""

    def __init__(self, measurement: Measurement, parent: QtCore.QObject = None) -> None:
        super().__init__(parent)
        self.m = measurement
        self.display_update_timer = QtCore.QTimer()
        self.display_update_timer.timeout.connect(self._on_display_update_timer)

    @QtCore.Slot()
    def _on_display_update_timer(self):
        try:
            self.m.update_display()
        except Exception as err:
            tb = "\n".join(traceback.format_exception(*sys.exc_info()))
            self.m.log.error(f"{self.m.name} failed to update display: {err}. {tb}")
        finally:
            if not self.m.is_measuring():
                self.display_update_timer.stop()


def to_etr_str(secs):
    mins, mins_rem = divmod(secs, 60)
    if not mins:
        return f"{int(secs)}s"
    hours, hours_rem = divmod(secs, 3600)
    if not hours:
        return f"{mins + mins_rem/60:0.1f}min"
    days, days_rem = divmod(secs, 86400)
    if not days:
        return f"{hours + hours_rem/3600:0.1f}h"
    return f"{secs/86400:0.1f}d"



"""
Created on $DATE_PRETTY

@author: $AUTHORS
"""

import pyqtgraph as pg
from qtpy import QtCore, QtWidgets

from ScopeFoundry import Measurement, h5_io


class $READOUT_CLASS_NAME(Measurement):

    name = "$READOUT_NAME"

    def setup(self):
        """
        Runs once during app initialization.
        This is where you define your settings, 
        and set up data structures.
        """
        s = self.settings
        s.New("save_h5", bool, initial=True)

        # data structure of the measurement
        self.data = {"y": [2, 5, 2], "img": [[2, 4], [1, 3]]}

    def setup_figure(self):
        """
        Runs once during app initialization and is responsible
        for creating a QtWidgets.QWidget self.ui.
        """
        cb_layout = QtWidgets.QHBoxLayout()
        cb_layout.addWidget(self.new_start_stop_button())
        cb_layout.addWidget(
            self.settings.New_UI(
                exclude=("activation", "run_state", "profile", "progress")
            )
        )
        header_widget = QtWidgets.QWidget()
        header_layout = QtWidgets.QVBoxLayout(header_widget)
        header_layout.addLayout(cb_layout)

        # plot
        self.graphics_widget = pg.GraphicsLayoutWidget(border=(100, 100, 100))
        self.plot = self.graphics_widget.addPlot(title=self.name)
        self.plot_lines = {}
        self.plot_lines["y"] = self.plot.plot(pen="g")
        # adding an image item
        # self.img_widget = self.graphics_widget.addPlot()
        # self.img_item = pg.ImageItem()
        # self.img_widget.addItem(self.img_item)

        # hw controls (also in the tree)
        # hw = self.app.hardware["$HW_NAME"]
        # hw_ctr = hw.new_control_widgets()

        # ScopeFoundry assumes .ui is the main widget:
        self.ui = QtWidgets.QSplitter(QtCore.Qt.Orientation.Vertical)
        self.ui.addWidget(header_widget)
        self.ui.addWidget(self.graphics_widget)
        # self.ui.addWidget(hw_ctr)

    def update_display(self):
        self.plot_lines["y"].setData(self.data["y"])
        # self.img_item.setImage(self.data["img"])

    def run(self):
        """
        Runs when the measurement starts. Executes in a separate thread from the GUI.
        It should not update the graphical interface directly and should focus only
        on data acquisition.
        """

        print(self.__class__, "run method not implemented")

        # hw = self.app.hardware["$HW_NAME"]
        # self.data['y'] = hw.dev.read_data()

        if self.settings["save_h5"]:
            self.save_h5(data=self.data)


    # ---------------------------------------------------------------------------
    ## UNCOMMENT IF YOU HAVE SCOPEFOUNDRY 2.0 OR EARLIER
    # ---------------------------------------------------------------------------
    # def save_h5(self, data):
    #     """
    #     opens new a .h5 file, stores *data*, and closes the file.

    #     If *dataset_metadata* is None, Measurement.new_dataset_metadata() is called which updates
    #     Measurement.dataset_metadata
    #     """
    #     self.open_new_h5_file(data=data)
    #     self.close_h5_file()

    # def open_new_h5_file(self, data = None):
    #     """
    #     sets attributes:
    #       - Measurement.h5_file
    #       - Measurement.h5_meas_group
    #     returns Measurement.h5_meas_group
    #     """

    #     # there might be a file open already
    #     self.close_h5_file()

    #     from ScopeFoundry import h5_io

    #     self.h5_file = h5_io.h5_base_file(self.app, measurement=self)
    #     self.h5_meas_group = h5_io.h5_create_measurement_group(self, self.h5_file)

    #     if data:
    #         for name, value in data.items():
    #             self.h5_meas_group.create_dataset(name=name, data=value)

    #     return self.h5_meas_group 

    # def close_h5_file(self):
    #     if hasattr(self, "h5_file") and self.h5_file.id is not None:
    #         self.h5_file.close()



import unittest

from pyqtgraph.widgets.SpinBox import SpinBox
from qtpy import QtWidgets

from ScopeFoundry.base_app.base_app import BaseApp
from ScopeFoundry.widgets import MinMaxQSlider


class LQDisconnectFromWidgetTest(unittest.TestCase):

    def setUp(self):
        self.app = app = BaseApp([])
        self.settings = app.settings
        self.str = self.settings.New("str", str, initial="0")
        self.float = self.settings.New("float", float, vmin=-100, vmax=100)
        self.int = self.settings.New("int", int)
        self.bool = self.settings.New("bool", bool)
        self.choices = self.settings.New(
            "choices", int, choices=(("one", 1), ("two", 2), ("three", 3))
        )
        self._dir = self.settings.New("dir", "file", initial="test_dir", is_dir=True)

    def tearDown(self):
        # self.app.on_close()
        self.app.qtapp.exit()
        del self.app

    def test_qdouble_spinbox(self):
        wf = QtWidgets.QDoubleSpinBox()
        wi = QtWidgets.QDoubleSpinBox()

        self.float.connect_to_widget(wf)
        self.int.connect_to_widget(wi)

        self.float.update_value(1.0)
        self.int.update_value(1.0)

        self.float.disconnect_from_widget(wf)
        self.int.disconnect_from_widget(wi)

        # widget to lq
        wf.setValue(2.0)
        wi.setValue(2)

        self.assertEqual(self.float.val, 1)
        self.assertEqual(self.int.val, 1)

        # lq to widget
        self.float.update_value(3.0)
        self.int.update_value(3)

        self.assertEqual(wf.value(), 2.0)
        self.assertEqual(wi.value(), 2)


#     def test_pyqtgraph_spinbox(self):
#         wf = SpinBox()
#         wi = SpinBox()

#         wf.setValue(3.0)
#         wi.setValue(3)

#         # test on connection widget take lq value
#         self.float.connect_to_widget(wf)
#         self.int.connect_to_widget(wi)
#         self.assertEqual(wf.value(), 0.0)
#         self.assertEqual(wi.value(), 0)

#         # lq to widget
#         self.float.update_value(1.0)
#         self.int.update_value(1.0)
#         self.assertEqual(wf.value(), 1)
#         self.assertEqual(wf.value(), 1)

#         # widget to lq
#         wf.setValue(2)
#         wi.setValue(2)
#         self.assertEqual(self.float.value, 2)
#         self.assertEqual(self.int.value, 2)

#     def test_qlabel(self):
#         ws = QtWidgets.QLabel()
#         wf = QtWidgets.QLabel()
#         wi = QtWidgets.QLabel()

#         ws.setText("3")
#         wf.setText("3")
#         wi.setText("3")

#         # test on connection widget take lq value
#         self.str.connect_to_widget(ws)
#         self.float.connect_to_widget(wf)
#         self.int.connect_to_widget(wi)
#         self.assertEqual(ws.text(), "0")
#         self.assertEqual(wf.text(), "0")
#         self.assertEqual(wi.text(), "0")

#         # lq to widget
#         self.str.update_value("1")
#         self.float.update_value(1.1)
#         self.int.update_value(1)
#         self.assertEqual(ws.text(), "1")
#         self.assertEqual(wf.text(), "1.1")
#         self.assertEqual(wi.text(), "1")

#         # widget to lq
#         # NA

#     def test_qline_edit(self):
#         ws = QtWidgets.QLineEdit()
#         wf = QtWidgets.QLineEdit()
#         wi = QtWidgets.QLineEdit()

#         ws.setText("3")
#         wf.setText("3")
#         wi.setText("3")

#         # test on connection widget take lq value
#         self.str.connect_to_widget(ws)
#         self.float.connect_to_widget(wf)
#         self.int.connect_to_widget(wi)
#         self.assertEqual(ws.text(), "0")
#         self.assertEqual(wf.text(), "0")
#         self.assertEqual(wi.text(), "0")

#         # lq to widget
#         self.str.update_value("1")
#         self.float.update_value(1.1)
#         self.int.update_value(1)
#         self.assertEqual(ws.text(), "1")
#         self.assertEqual(wf.text(), "1.1")
#         self.assertEqual(wi.text(), "1")

#         # widget to lq
#         ws.setText("2")
#         wf.setText("2")
#         wi.setText("2")
#         ws.editingFinished.emit()
#         wf.editingFinished.emit()
#         wi.editingFinished.emit()
#         self.assertEqual(self.str.value, "2")
#         self.assertEqual(self.float.value, 2.0)
#         self.assertEqual(self.int.value, 2)

#     def test_q_plain_text_edit(self):

#         ws = QtWidgets.QPlainTextEdit()
#         ws.document().setPlainText("3")

#         # test on connection widget take lq value
#         self.str.connect_to_widget(ws)
#         self.assertEqual(ws.document().toPlainText(), "0")

#         # lq to widget
#         self.str.update_value("1")
#         self.assertEqual(ws.document().toPlainText(), "1")

#         # widget to lq
#         ws.document().setPlainText("2")
#         self.assertEqual(self.str.value, "2")

#     def test_check_box(self):
#         wb = QtWidgets.QCheckBox()

#         # print("checkbox1")
#         wb.setChecked(True)

#         # test on connection widget take lq value
#         self.bool.connect_to_widget(wb)
#         self.assertEqual(wb.isChecked(), False)

#         # lq to widget
#         # print("checkbox2")

#         self.bool.update_value(True)
#         self.assertEqual(wb.isChecked(), True)

#         # widget to lq
#         # print("checkbox3")

#         wb.setChecked(False)

#         # TODO The fact that emit is required controversial. Can one also check/uncheck without clicking?
#         wb.clicked.emit()
#         self.assertEqual(self.bool.value, False)

#         # print("checkbox4")
#         wb.setChecked(True)
#         wb.clicked.emit(True)
#         self.assertEqual(self.bool.value, True)

#     def test_qcombo_box(self):
#         wc = QtWidgets.QComboBox()

#         wc.clear()
#         wc.addItem("-1", -1)
#         wc.addItem("-2", -2)
#         wc.setCurrentIndex(1)

#         # test on connection widget take lq value
#         self.choices.connect_to_widget(wc)
#         self.assertEqual(wc.currentData(), 1)

#         ## potentially too strong of a test.
#         self.assertEqual(wc.currentIndex(), 0)

#         # lq to widget
#         self.choices.update_value(1)
#         self.assertEqual(wc.currentData(), 1)

#         # widget to lq
#         wc.setCurrentIndex(2)
#         self.assertEqual(self.choices.value, 3)

#     def test_qslider_int_non_overflow(self):
#         lq = self.settings.New("int2", int, vmin=-100, vmax=100)

#         wi = QtWidgets.QSlider()
#         wi.setValue(3)

#         # test on connection widget take lq value
#         lq.connect_to_widget(wi)
#         self.assertEqual(wi.value(), 0.0)

#         # lq to widget
#         lq.update_value(1)
#         self.assertEqual(wi.value(), 1)

#         # widget to lq
#         wi.setValue(2)
#         self.assertEqual(lq.value, 2)

#     def test_qslider_int_overflow(self):
#         # falls back to act like a float
#         wi = QtWidgets.QSlider()
#         wi.setValue(int(1000))

#         lq = self.int

#         # test on connection widget take lq value
#         lq.connect_to_widget(wi)
#         self.assertAlmostEqual(lq._transform_from_slider(wi.value()), 0.0)

#         # lq to widget
#         lq.update_value(int(1e11))
#         self.assertAlmostEqual(
#             lq._transform_from_slider(wi.value()) / 1e12, int(1e11) / 1e12
#         )

#         # widget to lq
#         wi.setValue(lq._transform_to_slider(2e11))
#         self.assertAlmostEqual(lq.value / 1e12, 2e11 / 1e12)

#     def test_qslider_float(self):
#         wf = QtWidgets.QSlider()
#         wf.setValue(3)

#         # test on connection widget take lq value
#         self.float.connect_to_widget(wf)
#         self.assertAlmostEqual(self.float._transform_from_slider(wf.value()), 0.0)

#         # lq to widget
#         self.float.update_value(1.0)
#         self.assertAlmostEqual(self.float._transform_from_slider(wf.value()), 1.0)

#         # widget to lq
#         wf.setValue(self.float._transform_to_slider(2))
#         self.assertAlmostEqual(self.float.value, 2)

#     def test_min_max_slider(self):
#         wf = MinMaxQSlider()
#         wi = MinMaxQSlider()

#         wf.update_value(3.0)
#         wi.update_value(3)

#         # test on connection widget take lq value
#         self.float.connect_to_widget(wf)
#         self.int.connect_to_widget(wi)
#         self.assertEqual(wf.val, 0.0)
#         self.assertEqual(wi.val, 0.0)

#         # lq to widget
#         self.float.update_value(1.0)
#         self.int.update_value(1)
#         self.assertEqual(wf.val, 1)
#         self.assertEqual(wi.val, 1)

#         # widget to lq
#         wf.update_value(2)
#         wi.update_value(2)
#         self.assertEqual(self.float.value, 2)
#         self.assertEqual(self.int.value, 2)

#     def test_connect_to_file_widget(self):
#         wdir = QtWidgets.QLineEdit()
#         browseButton = QtWidgets.QPushButton("...")

#         # test on connection widget take lq value
#         wdir.setText("3")
#         self._dir.connect_to_browse_widgets(wdir, browseButton)
#         self.assertEqual(wdir.text(), "test_dir")

#         # lq to widget
#         self._dir.update_value("1")
#         self.assertEqual(wdir.text(), "1")

#         # widget to lq
#         wdir.setText("2")
#         wdir.editingFinished.emit()
#         self.assertEqual(self._dir.value, "2")


# class MyBaseApp(BaseApp):

#     def __init__(self, argv, **kwargs):
#         super().__init__(argv, **kwargs)

#         lq = self.settings.New("test_lq", float, 33, "Hz", vmin=10, vmax=133)
#         # lq = self.settings.New("test_lq", int, 33, "Hz", vmin=10, vmax=133)

#         # w = MinMaxQSlider("test")
#         w = QtWidgets.QSlider()
#         lq.connect_to_widget(w)
#         s = self.settings.New_UI(title="settings")

#         self.ui = QtWidgets.QWidget()
#         layout = QtWidgets.QVBoxLayout(self.ui)
#         layout.addWidget(w)
#         layout.addWidget(s)
#         self.ui.show()


if __name__ == "__main__":
    import sys

    # app = MyBaseApp([])
    # sys.exit(app.exec_())
    unittest.main()



import logging

import numpy as np
import pyqtgraph as pg

try:
    import IPython

    if (
        IPython.version_info[0] < 4
    ):  # compatibility for IPython < 4.0 (pre Jupyter split)
        from IPython.qt.console.rich_ipython_widget import (
            RichIPythonWidget as RichJupyterWidget,
        )
        from IPython.qt.inprocess import QtInProcessKernelManager
    else:
        from qtconsole.inprocess import QtInProcessKernelManager
        from qtconsole.rich_jupyter_widget import RichJupyterWidget
    CONSOLE_TYPE = "qtconsole"
except Exception as err:
    logging.warning(
        f"ScopeFoundry unable to import iPython console, using pyqtgraph.console instead. Error: {err}"
    )
    import pyqtgraph.console

    CONSOLE_TYPE = "pyqtgraph.console"


def new_console_widget(self, kernel=None):
    """
    Create and return console QWidget. If Jupyter / IPython is installed
    this widget will be a full-featured IPython console. If Jupyter is unavailable
    it will fallback to a pyqtgraph.console.ConsoleWidget.

    If the app is started in an Jupyter notebook, the console will be
    connected to the notebook's IPython kernel.

    the returned console_widget will also be accessible as console_widget

    In order to see the console widget, remember to insert it into an existing
    window or call console_widget.show() to create a new window
    """
    if CONSOLE_TYPE == "pyqtgraph.console":
        console_widget = pyqtgraph.console.ConsoleWidget(
            namespace={"app": self, "pg": pg, "np": np}, text="ScopeFoundry Console"
        )
    elif CONSOLE_TYPE == "qtconsole":

        if kernel is None:
            try:  # try to find an existing kernel
                # https://github.com/jupyter/notebook/blob/master/docs/source/examples/Notebook/Connecting%20with%20the%20Qt%20Console.ipynb
                import ipykernel as kernel

                conn_file = kernel.get_connection_file()
                import qtconsole.qtconsoleapp

                app = qtconsole.qtconsoleapp.JupyterQtConsoleApp()
                console_widget = app.new_frontend_connection(conn_file)
                console_widget.setWindowTitle("ScopeFoundry IPython Console")
            except:  # make your own new in-process kernel
                # https://github.com/ipython/ipython-in-depth/blob/master/examples/Embedding/inprocess_qtconsole.py
                kernel_manager = QtInProcessKernelManager()
                kernel_manager.start_kernel()
                kernel = kernel_manager.kernel
                kernel.shell.banner1 += """
                ScopeFoundry Console
                
                Variables:
                    * np: numpy package
                    * app: the ScopeFoundry App object
                """
                kernel.gui = "qt4"
                kernel.shell.push({"np": np, "app": self})
                kernel_client = kernel_manager.client()
                kernel_client.start_channels()

                # console_widget = RichIPythonWidget()
                console_widget = RichJupyterWidget()
                console_widget.setWindowTitle("ScopeFoundry IPython Console")
                console_widget.kernel_manager = kernel_manager
                console_widget.kernel_client = kernel_client
        else:
            import qtconsole.qtconsoleapp

            app = qtconsole.qtconsoleapp.JupyterQtConsoleApp()
            console_widget = app.new_frontend_connection(kernel.get_connection_file())
            console_widget.setWindowTitle("ScopeFoundry IPython Console")
    else:
        raise ValueError("CONSOLE_TYPE undefined")

    return console_widget



from typing import Sequence
from qtpy import QtWidgets

from ScopeFoundry.dynamical_widgets.generic_widget import add_to_layout
from ScopeFoundry.logged_quantity.collection import LQCollection
from ScopeFoundry.operations import Operations


class Page:
    name = None

    def __init__(self, app) -> None:
        if self.name is None:
            self.name = "page"

        self.settings = LQCollection()
        self.operations = Operations()
        self.procedure = ()
        self.app = app

    def setup(self):
        """override me"""
        self.procedure = ()
        self.name = "page"

    def setup_figure(self):
        self.ui = QtWidgets.QWidget()
        self.layout = QtWidgets.QFormLayout(self.ui)
        if not self.procedure:
            add_to_layout(self.settings, self.layout)
            add_to_layout(self.operations, self.layout)
            return
        procedure_to_layout(self.procedure, self.layout, self.settings, self.operations)


def procedure_to_layout(
    procedure: Sequence[str],
    layout: QtWidgets.QFormLayout,
    settings: LQCollection,
    operations: Operations,
):
    n_text_instruction = 1
    for step in procedure:
        if step in operations:
            layout.addRow(operations.new_button(step))
        elif step in settings:
            lq = settings.get_lq(step)
            layout.addRow(step, lq.new_default_widget())
        else:
            inst = QtWidgets.QLabel(f"{n_text_instruction}. {step}")
            inst.setStyleSheet(
                "font: italic bolt 18; padding-top:23px; padding-bottom: 10px"
            )
            layout.addRow(inst)
            n_text_instruction += 1



from pathlib import Path
from typing import Dict, List, Set, Tuple

import h5py

LOADERS_FNAME = "h5_data_loaders.py"

LOADERS_PY_HEADER = """# generated with ScopeFoundry.tools
#
# pip install ScopeFoundry
# python -m ScopeFoundry.tools
#
# from h5_data_loaders import load, find_settings
# data = load('your_file_name.h5')

import functools
import fnmatch
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict

import h5py
import numpy as np

load_funcs = {}


def get_measurement_name(fname: str) -> str:
    with h5py.File(fname, "r") as file:
        if len(file["measurement"].keys()) == 1:
            return list(file["measurement"].keys())[0]
        return file.attrs["measurement"]


def load(fname: str) -> Any:
    mm_name = get_measurement_name(fname)
    return load_funcs[mm_name](fname)


def load_settings(fname: str) -> Dict[str, Any]:
    path = Path(fname)
    if not path.suffix == ".h5":
        return {}

    settings = {}
    visit_func = functools.partial(_settings_visitfunc, settings=settings)

    with h5py.File(fname, "r") as file:
        for key, val in file.attrs.items():
            settings[key] = val       
        file.visititems(visit_func)
 
    return settings


def find_settings(settings: Dict, pattern: str = "measurement/*") -> Dict:
    if type(settings) is str:
        settings = load_settings(settings)
    if hasattr(settings, "settings"):
        settings = settings.settings
    matching = [key for key in settings.keys() if fnmatch.fnmatch(key, pattern)]
    if not matching:
        print("no matching key found")
        return {}
    return {k: settings[k] for k in matching}


def _settings_visitfunc(name: str, node: h5py.Group, settings: Dict[str, Any]) -> None:
    if not name.endswith("settings"):
        return

    for key, val in node.attrs.items():
        lq_path = f"{name.replace('settings', key)}"
        settings[lq_path] = val


def get_mm_name(fname: str) -> str:
    with h5py.File(fname, "r") as file:
        if len(file["measurement"].keys()) == 1:
            return list(file["measurement"].keys())[0]
        return file.attrs["measurement"]
"""
__STR__INFO = r"""

    def __str__(self) -> str:
        lines = [f"{self.path}"]
        for key, value in self.__dict__.items():
            if isinstance(value, np.ndarray):
                lines.append(f"-D {key}: {value.shape}")
        lines += [f"{key}: {value}" for key, value in self.settings.items()]
        return "\n".join(lines)
"""


def get_measurement_name(fname: str) -> str:
    with h5py.File(fname, "r") as file:
        if len(file["measurement"].keys()) == 1:
            return list(file["measurement"].keys())[0]
        return file.attrs["measurement"]


def generate_loaders(dsets: Dict[str, Set[str]]) -> List[str]:
    lines = []
    for mm_name, key_set in dsets.items():

        class_name = "".join(x.title() for x in mm_name.split("_"))

        data_class_lines = [
            "@dataclass",
            f"class {class_name}:",
            f"{' ':>4}path: Path",
            f"{' ':>4}settings: dict",
        ]
        load_func_lines = [
            f"def load_{mm_name}(fname: str) -> {class_name}:",
            f"{' ':>4}with h5py.File(fname, 'r') as file:",
            f"{' ':>8}m = file['measurement/{mm_name}']",
            f"{' ':>8}return {class_name}(",
            f"{' ':>12}path=Path(fname),",
            f"{' ':>12}settings=load_settings(fname),",
        ]
        for name in key_set:
            data_class_lines.append(f"{' ':>4}{name}: np.ndarray")
            load_func_lines.append(
                f"{' ':>12}{name}=m['{name}'][:] if '{name}' in m else None,"
            )

        data_class_lines.append(__STR__INFO)

        load_func_lines.append(f"{' ':>8})")

        lines.append("")
        lines += data_class_lines
        lines.append("")
        lines.append("")
        lines += load_func_lines
        lines.append("")
        lines.append("")
        lines.append(f"load_funcs['{mm_name}'] = load_{mm_name}")
        lines.append("")

    return lines


def get_dset_names(folder: str) -> Dict[str, Set[str]]:
    path = Path(folder)
    dset_names = {}
    for fname in path.rglob("*.h5"):
        mm_name = get_measurement_name(fname)
        with h5py.File(fname, "r") as file:
            new_keys = set(
                [
                    name
                    for name, val in file[f"measurement/{mm_name}"].items()
                    if isinstance(val, h5py.Dataset)
                ]
            )
            if mm_name in dset_names:
                dset_names[mm_name] = dset_names[mm_name].union(new_keys)
            else:
                dset_names[mm_name] = new_keys

    return dset_names


def generate_loaders_py(folder: str = ".") -> Tuple[Path, Dict[str, Set[str]]]:
    path = Path(folder)
    fnames = tuple(path.rglob("*.h5"))
    lines = [LOADERS_PY_HEADER]

    if len(fnames):
        dset_names = get_dset_names(folder)
        lines += generate_loaders(dset_names)
    else:
        print(f"WARNING no h5 files found in {folder}. No loaders created.")
        return Path(), {}

    loaders_fname = path / LOADERS_FNAME
    with open(loaders_fname, "w") as file:
        file.write("\n".join(lines))

    return loaders_fname, dset_names



import os
import sys
import webbrowser

from qtpy.QtWidgets import (
    QApplication,
    QFileDialog,
    QHBoxLayout,
    QLabel,
    QListWidget,
    QMessageBox,
    QPushButton,
    QVBoxLayout,
    QWidget,
)

from ScopeFoundry.tools.features.publish_on_gh.publish_all_on_gh import publish_on_gh
from ScopeFoundry.tools.page import Page


class PublishOnGH(Page):

    name = "publish on gh"

    def setup(self) -> None:
        self.name = "publish on gh"
        self.settings.new_file(
            "ScopeFoundryHW_dir", str, initial="/ScopeFoundryHW", is_dir=True
        )
        self.settings.New("gh_username", str, initial="ubene")
        self.settings.New("message", str, initial="Initial commit")
        self.settings.New(
            "private_or_public",
            str,
            choices=[("private", "--private"), ("public", "--public")],
            initial="--public",
            description="Choose whether to publish the repository as private or public.",
        )

    def setup_figure(self):
        self.ui = FolderListWidget(self.settings)


from ScopeFoundry.logged_quantity.collection import LQCollection

INSTRUCTIONS_URL = "https://scopefoundry.org/docs/200_hardware-sharing/"


class FolderListWidget(QWidget):

    @property
    def parent_dir(self):
        return self.settings["ScopeFoundryHW_dir"]

    def __init__(self, settings: LQCollection):
        super().__init__()
        self.settings = settings

        # Buttons
        self.instructions_button = QPushButton("Instructions")  # Instructions button

        # Widgets
        root_layout = QHBoxLayout()
        self.current_dir_widget = self.settings.get_lq(
            "ScopeFoundryHW_dir"
        ).new_default_widget()
        # self.select_folder_button = QPushButton("Select ScopeFoundryHW Folder")
        root_layout.addWidget(self.current_dir_widget)
        # root_layout.addWidget(self.select_folder_button)

        # list layout
        self.available_label = QLabel("Available HWs")
        self.selected_label = QLabel("HWs to publish")
        self.available_list = QListWidget()
        self.selected_list = QListWidget()

        left_list_layout = QVBoxLayout()
        right_list_layout = QVBoxLayout()
        left_list_layout.addWidget(self.available_label)
        right_list_layout.addWidget(self.selected_label)
        left_list_layout.addWidget(self.available_list)
        right_list_layout.addWidget(self.selected_list)
        list_layout = QHBoxLayout()
        list_layout.addLayout(left_list_layout)
        list_layout.addLayout(right_list_layout)

        self.move_to_selected_button = QPushButton("Move selected -->")
        self.move_to_available_button = QPushButton("Move selected <--")
        left_list_layout.addWidget(self.move_to_selected_button)
        right_list_layout.addWidget(self.move_to_available_button)

        self.publish_button = QPushButton("Publish!")  # Publish button

        # Connect buttons
        self.instructions_button.clicked.connect(self.open_instructions)
        self.settings.get_lq("ScopeFoundryHW_dir").add_listener(
            self.populate_available_list
        )
        self.move_to_selected_button.clicked.connect(self.move_to_selected)
        self.move_to_available_button.clicked.connect(self.move_to_available)
        self.publish_button.clicked.connect(self.publish_selected)

        # Publish layout
        publish_layout = QVBoxLayout()
        publish_widget = self.settings.New_UI(
            ("gh_username", "message", "private_or_public")
        )
        publish_layout.addWidget(publish_widget)
        publish_layout.addWidget(self.publish_button)

        # Add to main layout
        layout = QVBoxLayout()
        layout.addWidget(self.instructions_button)  # Add Instructions button
        layout.addLayout(root_layout)
        layout.addLayout(list_layout)
        layout.addLayout(publish_layout)  # Add Publish button to layout

        self.setLayout(layout)
        self.setWindowTitle("Publish HW layout")

        self.populate_available_list()

    def populate_available_list(self):
        """Populates the available list with subfolders in the parent directory."""
        self.available_list.clear()
        if not os.path.isdir(self.parent_dir):
            QMessageBox.critical(
                self, "Error", f"{self.parent_dir} is not a valid directory."
            )
            return

        for subfolder in os.listdir(self.parent_dir):
            subfolder_path = os.path.join(self.parent_dir, subfolder)
            if os.path.isdir(subfolder_path):
                self.available_list.addItem(subfolder)

    def move_to_selected(self):
        """Moves selected items from the available list to the selected list."""
        for item in self.available_list.selectedItems():
            self.selected_list.addItem(item.text())
            self.available_list.takeItem(self.available_list.row(item))

    def move_to_available(self):
        """Moves selected items from the selected list back to the available list."""
        for item in self.selected_list.selectedItems():
            self.available_list.addItem(item.text())
            self.selected_list.takeItem(self.selected_list.row(item))

    def publish_selected(self):
        """Callback for the Publish button. Add your publishing logic here."""
        selected_folders = [
            self.selected_list.item(i).text() for i in range(self.selected_list.count())
        ]
        # print(f"Publishing the following folders: {selected_folders}")

        for subfolder in selected_folders:
            subfolder_path = os.path.join(self.parent_dir, subfolder)

            # Skip if not a directory
            if not os.path.isdir(subfolder_path):
                continue
            print(f"Publishing {subfolder_path}")

            publish_on_gh(
                self.settings["gh_username"],
                subfolder,
                subfolder_path,
                message=self.settings["message"],
                private_or_public=self.settings["private_or_public"],
            )

        # Add your publishing logic here

    def open_instructions(self):
        """Opens the instructions webpage in the default browser."""
        webbrowser.open(INSTRUCTIONS_URL)

    def select_folder(self):
        """Opens a dialog to select a folder and updates the parent directory."""
        folder = QFileDialog.getExistingDirectory(
            self, "Select Folder", self.parent_dir
        )
        if folder:
            self.settings["ScopeFoundryHW_dir"] = folder
            self.current_dir_widget.setText(f"Current Directory: {self.parent_dir}")
            self.clear_lists()
            self.populate_available_list()

    def clear_lists(self):
        """Clears both lists."""
        self.available_list.clear()
        self.selected_list.clear()


if __name__ == "__main__":
    app = QApplication(sys.argv)

    # Replace with the path to your parent directory
    parent_directory = "/Users/benediktursprung/Library/CloudStorage/OneDrive-Personal/scope_foundries/scopefoundry_trovatello_lab/ScopeFoundryHW2"

    settings = LQCollection()
    settings.new_file("ScopeFoundryHW_dir", initial="/ScopeFoundryHW", is_dir=True)
    settings.New("gh_username", str, initial="ubene")
    settings.New("message", str, initial="Initial commit")
    settings.New(
        "private_or_public",
        str,
        choices=[("private", "--private"), ("public", "--public")],
        initial="--public",
        description="Choose whether to publish the repository as private or public.",
    )
    widget = FolderListWidget(settings=settings)
    widget.show()

    sys.exit(app.exec_())



from ScopeFoundry import HardwareComponent
from ScopeFoundry.examples.ScopeFoundryHW.bsinc_noiser200.bsinc_noiser200_dev import (
    Noiser200Dev,
)


class Noiser200HW(HardwareComponent):

    name = "noiser_200"

    def setup(self):

        self.settings.New("port", str, initial="Com200")
        self.signal = self.settings.New("signal", float, ro=True, unit="Hz", si=True)
        self.voltage = self.settings.New(
            name="voltage",
            dtype=float,
            unit="V",
            si=True,
            protected=True,  # do not want to write a voltage from a file
        )

    def connect(self):
        self.dev = Noiser200Dev(self.settings["port"])

        self.signal.connect_to_hardware(read_func=self.dev.read_signal)
        self.voltage.connect_to_hardware(write_func=self.dev.write_voltage)

    def disconnect(self):

        self.settings.disconnect_all_from_hardware()

        if hasattr(self, "dev"):
            del self.dev

    # def threaded_update(self):  # define for continueous update after connect
    #     self.signal.read_from_hardware()
    #     time.sleep(0.1)



import json
import sys
import threading
import time
import warnings
from functools import partial
from pathlib import Path
from typing import Callable

from qtpy import QtCore, QtGui, QtWidgets

from .base_app import BaseMicroscopeApp
from .dynamical_widgets import add_to_layout, new_widget
from .dynamical_widgets.tree_widget import SubtreeManager
from .helper_funcs import (
    QLock,
    get_child_path,
    get_logger_from_class,
    init_docs_path,
    itemize_launchers,
)
from .logged_quantity import LQCollection
from .operations import Operations


class HardwareComponent:
    """
    :class:`HardwareComponent`

    Base class for ScopeFoundry Hardware objects

    to subclass, implement :meth:`setup`, :meth:`connect` and :meth:`disconnect`

    """

    def __init__(self, app: BaseMicroscopeApp, debug: bool = False, name: str = None):
        """
        create new HardwareComponent attached to *app*
        """
        if not hasattr(self, "name"):
            self.name = self.__class__.__name__

        if name is not None:
            self.name = name

        self.settings = LQCollection(path=f"hw/{self.name}")
        self.operations = Operations(path=f"hw/{self.name}")
        self._subtree_managers_ = []
        self._widgets_managers_ = []

        self.log = get_logger_from_class(self)
        self.docs_path = get_child_path(self) / "docs"

        # threading lock
        # self.lock = threading.Lock()
        # self.lock = DummyLock()
        self.lock = QLock(mode=1)  # mode 0 is non-reentrant lock

        self.app = app

        self.connected = self.settings.New(
            "connected",
            dtype=bool,
            colors=["none", "rgba( 0, 255, 0, 80)"],
            description=f"to <i>{self.name}</i> hardware if checked.",
        )

        self.debug_mode = self.settings.New(
            "debug_mode", dtype=bool, initial=debug, colors=["none", "yellow"]
        )

        # self.connect_success = False # ever used?
        self.auto_thread_lock = True

        self.setup()

        if self.auto_thread_lock:
            self.thread_lock_all_lq()

        # self.has_been_connected_once = False # ever used?
        # self.is_connected = False # ever used?

        self.q_object = HardwareQObject()
        self.connection_succeeded = self.q_object.connection_succeeded
        self.connection_failed = self.q_object.connection_failed

        self.add_operation("Reload Code", self.reload_code)
        self.add_operation("Read from\nHardware", self.read_from_hardware)

        self.q_object.connection_failed.connect(self.on_connection_failed)
        self.q_object.connection_succeeded.connect(self.on_connection_succeeded)
        self.connected.updated_value[bool].connect(self.enable_connection)

    def enable_connection(self, enable=True):
        if enable:
            try:
                self.connect()
                # start thread if needed
                if hasattr(self, "run"):
                    self.update_thread_interrupted = False
                    self._update_thread = threading.Thread(target=self.run)
                    self._update_thread.start()

                self.connection_succeeded.emit()
            except Exception as err:
                self.connection_failed.emit()
                raise err
        else:
            print("disabling connection")
            try:
                try:
                    if hasattr(self, "run") and hasattr(self, "_update_thread"):
                        self.update_thread_interrupted = True
                        self._update_thread.join(timeout=5.0)
                        del self._update_thread
                finally:
                    self.disconnect()
                    self.set_connection_status("", "orange")
            except Exception as err:
                self.set_connection_status("⚠", "red")
                raise err

    def run(self):
        if hasattr(self, "threaded_update"):
            while not self.update_thread_interrupted:
                try:
                    self.threaded_update()
                except Exception as err:
                    print("threaded update failed", err)
                    time.sleep(1.0)

    def read_from_hardware(self):
        """
        Read all settings (:class:`LoggedQuantity`) connected to hardware states
        """
        for name, lq in self.settings.as_dict().items():
            if lq.has_hardware_read():
                lq.read_from_hardware()
                if self.debug_mode.val:
                    self.log.debug(f"read_from_hardware {name}: {lq.val}")

    def add_logged_quantity(self, name, **kwargs):
        return self.settings.New(name, **kwargs)

    def on_connection_succeeded(self):
        print(self.name, "connection succeeded!")
        self.connected.update_value(True)
        self.set_connection_status("✓", "green")

    def on_connection_failed(self):
        print(self.name, "connection failed!")
        self.connected.update_value(False)
        self.set_connection_status("⚠", "red")

    @property
    def gui(self):
        warnings.warn(
            "Hardware.gui is deprecated, use Hardware.app", DeprecationWarning
        )
        return self.app

    def web_ui(self):
        return f"Hardware {self.name}"

    def thread_lock_lq(self, lq):
        lq.old_lock = lq.lock
        lq.lock = self.lock

    def thread_lock_all_lq(self):
        for lq in self.settings.as_list():
            lq.old_lock = lq.lock
            lq.lock = self.lock

    def reload_code(self):
        import inspect

        if sys.version_info[1] <= 11:
            import xreload
        else:
            from ScopeFoundry import xreload

        mod = inspect.getmodule(self)

        x = xreload.xreload(mod)
        print("Reloaded code", x)

    def New_UI(self):
        scroll_area = self.settings.New_UI(style="scroll_form")
        for name in self.operations:
            btn = self.operations.new_button(name)
            scroll_area.widget().layout().addRow(btn)
        return scroll_area

    def new_control_widgets(
        self, title: str = None, include=None, exclude=None, style="scroll_form"
    ):
        """creates scroll area group box that updates on dynamical add/remove of settings/operations"""
        if title is None:
            title = self.name
        return new_widget(self, title, include, exclude, style)

    def add_to_layout(self, layout, include=None, exclude=None):
        add_to_layout(self, layout, include, exclude)

    def add_operation(
        self, name: str, op_func: Callable[[], None], description="", icon_path=""
    ):
        """
        Create an operation for the HardwareComponent.

        *op_func* is a function that will be called upon operation activation

        operations are typically exposed in the default ScopeFoundry gui via a pushButton

        :type name: str
        :type op_func: QtCore.Slot or Callable without Argument
        :type description: str
        """
        self.operations.new(name, op_func, description, icon_path)

    def remove_operation(self, name):
        self.operations.remove(name)

    def on_new_subtree(self, subtree: SubtreeManager):
        status_text = QtWidgets.QLabel()

        connect_cb = QtWidgets.QCheckBox()
        s0 = connect_cb.styleSheet()
        self.settings.get_lq("connected").connect_to_widget(connect_cb)
        connect_cb.setStyleSheet(s0)
        connect_cb.setFixedWidth(22)
        connect_cb.setFixedHeight(22)
        connect_cb.setToolTip(f"connect/disconnect {self.name}")

        widget = QtWidgets.QWidget()
        layout = QtWidgets.QHBoxLayout(widget)
        layout.addWidget(status_text)
        layout.addWidget(connect_cb)
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)
        subtree.tree_widget.setItemWidget(subtree.header_item, 1, widget)
        subtree.status_text = status_text

    def set_connection_status(self, text, color):
        for manager in self._subtree_managers_:
            manager.status_text.setText(text)
            manager.status_text.setStyleSheet(f"QLabel {{color : {color}; }}")

    def on_right_click(self):
        cmenu = QtWidgets.QMenu()
        a = cmenu.addAction(self.name)
        a.setEnabled(False)

        connect = partial(self.settings.get_lq("connected").update_value, True)
        disconnect = partial(self.settings.get_lq("connected").update_value, False)

        cmenu.addAction("Connect", connect)
        cmenu.addAction("Disconnect", disconnect)

        init_docs_path(self.docs_path, self.settings)
        pairs = itemize_launchers(self.docs_path, self.app.launch_browser)
        if pairs:
            cmenu.addSeparator()
            for name, func in pairs:
                cmenu.addAction(name, func)

        cmenu.exec_(QtGui.QCursor.pos())

    def setup(self):
        """
        Runs during __init__, before the hardware connection is established
        Should generate desired LoggedQuantities, operations
        """
        raise NotImplementedError()

    def connect(self):
        """
        Opens a connection to hardware
        and connects :class:`LoggedQuantity` settings to related hardware
        functions and parameters
        """
        raise NotImplementedError()

    def disconnect(self):
        """
        Disconnects the hardware and severs hardware--:class:`LoggedQuantity` links
        """

        raise NotImplementedError()


class HardwareQObject(QtCore.QObject):

    connection_succeeded = QtCore.Signal()
    connection_failed = QtCore.Signal()



# import os

# os.environ["QT_API"] = "pyqt6"
import time
from typing import List

import numpy as np
from qtpy import QtCore, QtWidgets

from ScopeFoundry import (
    BaseApp,
    BaseMicroscopeApp,
    HardwareComponent,
    Measurement,
    new_tree_widget,
)


class LQWidgetTestApp(BaseApp):

    name = "LQWidgetTestApp"

    def __init__(self, argv: List[str]) -> None:
        BaseApp.__init__(self, argv)

        self.settings.New(
            "long_string_test",
            dtype=str,
            initial="Lorem ipsum dolor sit amet, consectetur adipiscing elit. Aliquam accumsan, ligula a tristique luctus, felis est blandit libero, nec placerat justo diam at lacus. Fusce volutpat vitae lacus non lobortis. Fusce porttitor varius placerat. Curabitur et varius urna, sit amet gravida leo. Etiam eleifend luctus erat, vel maximus libero lacinia at. Pellentesque mattis pulvinar sem, sit amet porttitor mi maximus in. Sed venenatis orci sit amet nulla luctus, vitae pulvinar neque facilisis. Donec in felis sodales libero fringilla aliquam eu non urna. Praesent ac elit ac lorem cursus aliquam eu venenatis mauris. Proin sed aliquet nunc. Duis venenatis mi dapibus.",
            ro=False,
        )

        self.ui = self.settings.New_UI()

        long_string_test_plainTextEdit = QtWidgets.QPlainTextEdit()
        self.ui.layout().addRow(
            "long_string_test_plainTextEdit", long_string_test_plainTextEdit
        )

        self.settings.long_string_test.connect_to_widget(long_string_test_plainTextEdit)

        self.console_widget.show()
        self.ui.show()


class Hardware(HardwareComponent):

    name = "hardware"

    def setup(self) -> None:
        self.settings.New("val1", initial=0.1)
        self.settings.New("val2", str, initial="hello world")
        self.settings.New("val3", str, choices=("Apple", "Bear", "Car"))

    def connect(self) -> None:
        self.settings.New("val4", int, 33).connect_to_hardware(lambda: 33, print)
        print("connected to", self.name)

    def disconnect(self) -> None:

        self.settings.remove("val4")
        print("disconnected from", self.name)


class Measure(Measurement):

    name = "measure"

    def setup(self) -> None:

        self.settings.New(
            "color test",
            str,
            choices=("yellow", "blue", "grey"),
            colors=("yellow", "blue"),
        )
        self.settings.New("run_crash_immediately", dtype=bool, initial=False)
        self.settings.New("run_crash_middle", dtype=bool, initial=False)

        self.settings.New("name", dtype=str, initial="cool_op_name")

        self.add_operation("add_setting", self.on_add_setting)
        self.add_operation("remove_setting", self.on_remove_setting)

        self.add_operation("add_operation", self.on_add_operation)
        self.add_operation("remove_operation", self.on_remove_operation)

        self.data_array = np.zeros(100)

    def on_add_setting(self) -> None:
        self.settings.New(self.settings["name"], str, "generated")

    def on_remove_setting(self) -> None:
        self.settings.remove(self.settings["name"])

    def on_add_operation(self) -> None:
        self.add_operation(
            self.settings["name"],
            lambda: print("operation called"),
        )
        print(list(self.operations.keys()))

    def on_remove_operation(self) -> None:
        self.remove_operation(self.settings["name"])
        print(list(self.operations.keys()))

    def setup_figure(self) -> None:

        splitter = QtWidgets.QSplitter(QtCore.Qt.Orientation.Horizontal)
        # splitter.addWidget(self.app.hardware["hardware"].New_UI())
        splitter.addWidget(
            self.app.hardware["hardware"].new_control_widgets(
                include=("val*", "relo*"),
                title="incl: val*,relo*  excl: *1",
                exclude=("*1",),
            )
        )

        splitter.addWidget(self.new_control_widgets())
        widget = QtWidgets.QWidget()
        vbox_layout = QtWidgets.QVBoxLayout(widget)
        label = QtWidgets.QLabel(
            "used add_to_layout with a given QVBoxLayout to dyn add/remove widgets. \n include pattern matching: cool_*,add*, remove*, name"
        )
        vbox_layout.addWidget(label)
        splitter.addWidget(widget)
        self.add_to_layout(vbox_layout, include=("cool_*", "add*", "remove*", "name"))

        grid_widget = QtWidgets.QWidget()
        grid_layout = QtWidgets.QGridLayout(grid_widget)
        splitter.addWidget(grid_widget)
        self.add_to_layout(grid_layout, include=("cool_*", "add*", "remove*", "name"))

        self.ui = QtWidgets.QWidget()
        vsplitter = QtWidgets.QSplitter(QtCore.Qt.Orientation.Vertical)
        layout = QtWidgets.QVBoxLayout(self.ui)
        layout.addWidget(self.activation.new_pushButton())
        layout.addWidget(self.new_start_stop_button())
        layout.addWidget(vsplitter)
        vsplitter.addWidget(
            new_tree_widget(
                (self.app.measurements["measure"], self.app.hardware["hardware"]),
                ("hello", "world"),
            )
        )
        vsplitter.addWidget(splitter)

    def run(self) -> None:

        if self.settings["run_crash_immediately"]:
            raise IOError("run_crash_immediately")

        N = 10

        self.data_array = np.zeros(N)

        for i in range(N):
            # print(self.name, i, 'of', N)
            if self.interrupt_measurement_called:
                print(self.name, "interrupted at", i, "of", N)
                break
            self.set_progress(100.0 * i / N)
            if i > N / 2 and self.settings["run_crash_middle"]:
                raise IOError("run_crash_middle")

            self.settings.New(f"cool_{i}")
            self.add_operation(f"cool_{i}", lambda: print("hello"))
            time.sleep(0.1)

        for i in range(N):
            self.settings.remove(f"cool_{i}")
            self.remove_operation(f"cool_{i}")
            time.sleep(0.1)


class MeasureNonUI(Measurement):

    name = "measure_non_ui"

    def run(self) -> None:
        time.sleep(0.1)
        print(self.name, "run end")


class LQWidgetMicroscopeTestApp(BaseMicroscopeApp):

    name = "lq_widget_test_app"
    mdi = True

    def setup(self) -> None:

        self.add_hardware(Hardware(self))
        self.add_measurement(Measure(self))
        self.add_measurement(MeasureNonUI(self))


if __name__ == "__main__":
    import sys

    # app = LQWidgetTestApp(sys.argv)
    app = LQWidgetMicroscopeTestApp(sys.argv)
    app.exec_()



from typing import Iterable, List, Tuple

from ScopeFoundry.scanning.actuators import (
    ActuatorDefinitions,
    get_actuator_funcs,
    add_all_possible_actuators_and_parse_definitions,
)

from .base_raster_slow_scan import BaseRaster2DSlowScan, BaseRaster3DSlowScan


class BaseRaster2DSlowScanV2(BaseRaster2DSlowScan):

    name = "BaseRaster2DSlowScanV2"

    def __init__(
        self,
        app,
        name: str = None,
        actuators: Iterable[ActuatorDefinitions] = (),
        h_limits: Tuple[float, float] = (-1_000_000_000_000, 1_000_000_000_000),
        v_limits: Tuple[float, float] = (-1_000_000_000_000, 1_000_000_000_000),
        h_unit: str = "",
        v_unit: str = "",
        h_spinbox_decimals: int = 4,
        v_spinbox_decimals: int = 4,
        h_spinbox_step: float = 0.1,
        v_spinbox_step: float = 0.1,
        use_external_range_sync: bool = False,
        circ_roi_size: float = 1.0,
    ):
        self.actuators = actuators
        super().__init__(
            app=app,
            h_limits=h_limits,
            v_limits=v_limits,
            h_unit=h_unit,
            v_unit=v_unit,
            h_spinbox_decimals=h_spinbox_decimals,
            v_spinbox_decimals=v_spinbox_decimals,
            h_spinbox_step=h_spinbox_step,
            v_spinbox_step=v_spinbox_step,
            use_external_range_sync=use_external_range_sync,
            circ_roi_size=circ_roi_size,
        )
        if name is not None:
            self.name = name

    def setup(self):
        super().setup()
        self.h_actuator = self.settings.New("h_actuator", str, choices=())
        self.v_actuator = self.settings.New("v_actuator", str, choices=())
        self.update_actuator_choices()

        self.ui.h_axis_comboBox.setVisible(True)
        self.ui.v_axis_comboBox.setVisible(True)

        self.h_actuator.connect_to_widget(self.ui.h_axis_comboBox)
        self.v_actuator.connect_to_widget(self.ui.v_axis_comboBox)

        self.add_operation("update_actuator_choices", self.update_actuator_choices)

    def update_actuator_choices(self):
        self.actuator_defs = add_all_possible_actuators_and_parse_definitions(
            self.actuators, self.app
        )
        self.actuator_funcs = get_actuator_funcs(self.app, self.actuator_defs)
        choices = list(self.actuator_funcs.keys())
        self.h_actuator.change_choice_list(choices, choices[0])
        self.v_actuator.change_choice_list(choices, choices[1])

    def move_position_slow(self, h, v, dh, dv):
        self.move_position_fast(h, v, dh, dv)

    def move_position_fast(self, h, v, dh, dv):
        self.move_position_start(h, v)

    def move_position_start(self, h, v):
        h_read, h_write = self.actuator_funcs[self.settings["h_actuator"]]
        v_read, v_write = self.actuator_funcs[self.settings["v_actuator"]]

        h_write(h)
        v_write(v)

    def connect_pos_widgets(self):
        try:
            self.disconnect_pos_widgets()
        except ValueError as err:
            print(err)

        sh = self.app.get_lq(self.actuator_defs[self.settings["h_actuator"]][1])
        sv = self.app.get_lq(self.actuator_defs[self.settings["v_actuator"]][1])

        sh.connect_to_widget(self.ui.x_doubleSpinBox)
        sv.connect_to_widget(self.ui.y_doubleSpinBox)

        sh.add_listener(self.update_arrow_pos)
        sv.add_listener(self.update_arrow_pos)

    def disconnect_pos_widgets(self):
        sh = self.app.get_lq(self.actuator_defs[self.settings["h_actuator"]][1])
        sv = self.app.get_lq(self.actuator_defs[self.settings["v_actuator"]][1])

        sh.disconnect_from_widget(self.ui.x_doubleSpinBox)
        sv.disconnect_from_widget(self.ui.y_doubleSpinBox)

        sh.listeners.remove(self.update_arrow_pos)
        sv.listeners.remove(self.update_arrow_pos)

    def update_arrow_pos(self):
        rh, wh = self.actuator_defs[self.settings["h_actuator"]]
        rv, wv = self.actuator_defs[self.settings["v_actuator"]]

        h = self.app.get_lq(wh).value
        v = self.app.get_lq(wv).value

        self.current_stage_pos_arrow.setPos(h, v)


class BaseRaster3DSlowScanV2(BaseRaster3DSlowScan):

    name = "BaseRaster3DSlowScanV2"

    def __init__(
        self,
        app,
        name: str = None,
        actuators: List[ActuatorDefinitions] = (),
        h_limits: Tuple[float, float] = (-1_000_000_000_000, 1_000_000_000_000),
        v_limits: Tuple[float, float] = (-1_000_000_000_000, 1_000_000_000_000),
        z_limits: Tuple[float, float] = (-1_000_000_000_000, 1_000_000_000_000),
        h_unit: str = "",
        v_unit: str = "",
        z_unit: str = "",
        h_spinbox_decimals: int = 4,
        v_spinbox_decimals: int = 4,
        z_spinbox_decimals: int = 4,
        h_spinbox_step: float = 0.1,
        v_spinbox_step: float = 0.1,
        z_spinbox_step: float = 0.1,
        use_external_range_sync: bool = False,
        circ_roi_size: float = 1.0,
    ):
        self.actuators = actuators
        super().__init__(
            app=app,
            h_limits=h_limits,
            v_limits=v_limits,
            z_limits=z_limits,
            h_unit=h_unit,
            v_unit=v_unit,
            z_unit=z_unit,
            h_spinbox_decimals=h_spinbox_decimals,
            v_spinbox_decimals=v_spinbox_decimals,
            z_spinbox_decimals=z_spinbox_decimals,
            h_spinbox_step=h_spinbox_step,
            v_spinbox_step=v_spinbox_step,
            z_spinbox_step=z_spinbox_step,
            use_external_range_sync=use_external_range_sync,
            circ_roi_size=circ_roi_size,
        )
        if name is not None:
            self.name = name

    def setup(self):
        super().setup()
        self.h_actuator = self.settings.New("h_actuator", str, choices=())
        self.v_actuator = self.settings.New("v_actuator", str, choices=())
        self.z_actuator = self.settings.New("z_actuator", str, choices=())
        self.update_actuator_choices()

        self.ui.h_axis_comboBox.setVisible(True)
        self.ui.v_axis_comboBox.setVisible(True)
        self.ui.z_axis_comboBox.setVisible(True)

        self.h_actuator.connect_to_widget(self.ui.h_axis_comboBox)
        self.v_actuator.connect_to_widget(self.ui.v_axis_comboBox)
        self.z_actuator.connect_to_widget(self.ui.z_axis_comboBox)

    def update_actuator_choices(self):
        self.actuator_defs = add_all_possible_actuators_and_parse_definitions(
            self.actuators, self.app
        )
        self.actuator_funcs = get_actuator_funcs(self.app, self.actuator_defs)
        choices = list(self.actuator_funcs.keys())
        self.h_actuator.change_choice_list(choices, choices[0])
        self.v_actuator.change_choice_list(choices, choices[1])
        self.z_actuator.change_choice_list(choices, choices[2])

    def move_position_slow(self, h, v, dh, dv):
        h_read, h_write = self.actuator_funcs[self.settings["h_actuator"]]
        v_read, v_write = self.actuator_funcs[self.settings["v_actuator"]]

        h_write(h)
        v_write(v)

    def move_position_fast(self, h, v, dh, dv):
        self.move_position_slow(h, v, dh, dv)

    def move_position_start(self, h: float, v: float, z: float):
        h_read, h_write = self.actuator_funcs[self.settings["h_actuator"]]
        v_read, v_write = self.actuator_funcs[self.settings["v_actuator"]]
        z_read, z_write = self.actuator_funcs[self.settings["z_actuator"]]

        h_write(h)
        v_write(v)
        z_write(z)

    def connect_pos_widgets(self):
        try:
            self.disconnect_pos_widgets()
        except ValueError as err:
            print(err)

        sh = self.app.get_lq(self.actuator_defs[self.settings["h_actuator"]][1])
        sv = self.app.get_lq(self.actuator_defs[self.settings["v_actuator"]][1])
        sz = self.app.get_lq(self.actuator_defs[self.settings["z_actuator"]][1])

        sh.connect_to_widget(self.ui.x_doubleSpinBox)
        sv.connect_to_widget(self.ui.y_doubleSpinBox)
        sz.connect_to_widget(self.ui.z_doubleSpinBox)

        sh.add_listener(self.update_arrow_pos)
        sv.add_listener(self.update_arrow_pos)
        # sz.add_listener(self.update_arrow_pos)

    def disconnect_pos_widgets(self):

        sh = self.app.get_lq(self.actuator_defs[self.settings["h_actuator"]][1])
        sv = self.app.get_lq(self.actuator_defs[self.settings["v_actuator"]][1])
        sz = self.app.get_lq(self.actuator_defs[self.settings["z_actuator"]][1])

        sh.disconnect_from_widget(self.ui.x_doubleSpinBox)
        sv.disconnect_from_widget(self.ui.y_doubleSpinBox)
        sz.disconnect_from_widget(self.ui.z_doubleSpinBox)

        sh.listeners.remove(self.update_arrow_pos)
        sv.listeners.remove(self.update_arrow_pos)
        # sz.listeners.remove(self.update_arrow_pos)

    def update_arrow_pos(self):
        rh, wh = self.actuator_defs[self.settings["h_actuator"]]
        rv, wv = self.actuator_defs[self.settings["v_actuator"]]
        # rz, wz = self.actuators_definitions[self.settings["z_actuator"]]

        h = self.app.get_lq(wh).value
        v = self.app.get_lq(wv).value
        # z = self.app.get_lq(wz).value

        self.current_stage_pos_arrow.setPos(h, v)



from pathlib import Path
from typing import Dict

from ScopeFoundry.tools.features.utils import (
    mk_authors,
    mk_dates,
    to_class_name,
    write_templated_file,
)


def gather_infos(name: str, module_path: str = ".") -> Dict[str, str]:
    readout_file_name = f"{name.lower()}.py"
    readout_class_name = to_class_name(name)
    module_name = name
    module_path = Path(module_path.lower())
    readout_import_line = f"from {'.'.join(list(Path(module_path).parts)+[module_name.rstrip('.py')])} import {readout_class_name}"
    add_to_app = f"""
        {readout_import_line}
        self.add_measurement({readout_class_name}(self))
"""
    infos = {
        "READOUT_FILE_NAME": readout_file_name,
        "READOUT_CLASS_NAME": readout_class_name,
        "READOUT_NAME": f"{name.lower()}_readout",
        "TEST_APP_NAME": f"{name.lower()}_test_app",
        "HW_NAME": "",
        "MODULE_PATH": module_path,
        "READOUT_IMPORT_LINE": readout_import_line,
        "ADD_TO_APP": add_to_app,
    }
    infos.update(mk_dates())
    infos.update(mk_authors(""))
    return infos


def new_measure(name: str) -> Dict[str, str]:

    path_to_module = Path("measurements")
    if not path_to_module.exists():
        path_to_module.mkdir(parents=True)

    infos = gather_infos(name, str(path_to_module))
    root = Path(__file__).parent.parent

    write_templated_file(
        template_path=root / "templates/_readout.py",
        target_fname=infos["READOUT_FILE_NAME"],
        substitutions=infos,
        path_to_target=Path(infos["MODULE_PATH"]),
    )
    return infos


def main() -> None:
    name = input("name of measurement: ")
    subs = new_measure(name)

    print("add to your setup")
    print(subs["ADD_TO_APP"])


if __name__ == "__main__":
    main()
